from contextlib import contextmanager
from typing import Iterator, Any
from types import TracebackType

class Next:
	"""
	Context manager wrapper around the next(iter) function, to be used in
	when separating blocks of code by indent level would be useful
	"""
	slots=('el',)
	def __init__(self, iterator : Iterator[Any]):
		self.el = next(iterator)
	def __enter__(self):
		return(self.el)
	def __exit__(self, etype : type, evalue : Exception, traceback : TracebackType):
		return(False)



import numpy as np

import matplotlib.pyplot as plt

def keep(label_map, n_labels, to_keep_labels):
	for label in range(1, n_labels+1):
		if label not in to_keep_labels:
			label_map[label_map==label] = 0
	return label_map, to_keep_labels

def remove(label_map, to_remove_labels):
	for label in to_remove_labels:
		label_map[label_map == label] = 0
	return label_map, tuple(x for x in range(1, np.max(label_map)+1) if x not in to_remove_labels)

def relabel(label_map, n_old_labels, new_labels):
	for i, label in enumerate(new_labels):
		label_map[label_map == label] = n_old_labels + i + 1
	label_map[label_map > 0] -= n_old_labels
	return label_map, len(new_labels)

def keep_n_largest(label_map : np.ndarray, n_labels : int, n_to_keep : int):
	# order labels by number of pixels they contain
					
	ordered_labels = list(range(1,n_labels+1)) # label 0 is background
	ordered_labels.sort(key = lambda x: np.count_nonzero(label_map == x), reverse=True)
	
	if n_labels > n_to_keep:
		label_map, _ = keep(label_map, n_labels, ordered_labels[:n_to_keep])
		
		# relabel regions so they are in descending size order
		label_map, n_labels = relabel(label_map, n_labels, ordered_labels[:n_to_keep])
	
	return label_map, n_labels

def keep_larger_than(label_map : np.ndarray, n_labels : int, max_remove_size : int):
	kept_labels = []
	for i in range(1, n_labels+1):
		lbl_mask = label_map==i
		n_pixels = np.count_nonzero(lbl_mask)
		if n_pixels <= max_remove_size:
			label_map[lbl_mask] = 0 # set too-small regions to background
		else:
			kept_labels.append(i)
	
	label_map, n_labels = relabel(label_map, n_labels, kept_labels)
	return label_map, n_labels


"""
ND interpolation for numpy arrays
"""
from typing import TypeVar, NewType
import numpy as np


import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')

import aopp_deconv_tool.numpy_helper as nph
import aopp_deconv_tool.numpy_helper.array
import aopp_deconv_tool.numpy_helper.array.index

from aopp_deconv_tool.typedef import NumVar, ShapeVar

T = TypeVar('T')

type N = NewType('N', NumVar)
type M = NewType('M', NumVar)

type S[X] = NewType('ShapeS', ShapeVar[X])
type Q[X] = NewType('ShapeQ', ShapeVar[X])
type R[X] = NewType('ShapeR', ShapeVar[X])


def get_index_boundary_func(name : str):
	return {	
		'pacman' : nph.array.index.pacman,
		'periodic' : nph.array.index.periodic,
		'const' : nph.array.index.const,
		'reflect' : nph.array.index.reflect
	}[name]

def constant(a : np.ndarray[S[N],T], m : np.ndarray[S[N],bool], value : T = 0 ) -> np.ndarray[S[N],T]:
	"""
	Set masked elements to a constant value
	"""
	a[m] = value
	return a

def mean(
		a : np.ndarray[S[N],T], 
		m : np.ndarray[S[N],bool], 
		window : np.ndarray[Q[M],bool] | Q[M] | int = 3, # Where M <= N
		boundary : str ='reflect', 
		const : T = 0 
	) -> np.ndarray[S[N],T]:
	"""
	Replaces masked elements with mean of surrounding values in place.
	
	Arguments:
		a : np.ndarray[T]
			A numpy array of pixel data
		m : np.ndarray[bool]
			A mask of the elements of a to replace
		window : np.ndarray[bool] | tuple[int,...] | int
			The surrounding values we should use in the mean calculation. NANs
			and INFs will be ignored. If an `np.ndarray[bool]` will use the
			True elements as memmbers of the calculation, if a tuple will create
			a mask by using	each element of the tuple as a manhattan distance 
			from the point whose mean is being calculated, if an integer will
			use the same manhattan distance for each axis.
		boundary : str
			How boundaries are handled, one of ['pacman','const','reflect']
		const : T
			Value
	"""
	a = np.array(a)
	b_func = get_index_boundary_func(boundary)
	if boundary == 'const' : 
		index_boundary_func = lambda *args, **kwargs: b_func(*args, const = const, **kwargs)
	else:
		index_boundary_func = b_func
	
	idx_array = nph.array.indices_of_mask(m)
	if type(window) is np.ndarray:
		_lgr.debug(f'{a.ndim=} {window.ndim=}')
		while window.ndim < a.ndim:
			window = window[None,...]
		_lgr.debug(f'{window=}')
		deltas = nph.array.offsets_from_centre_of_mask(np.array(window, bool))
	else:
		deltas = nph.array.offsets_manhattan_distance(window, a.ndim)
	_lgr.debug(f'{deltas=}')
	
	n = idx_array[0].size
	contributions = np.zeros((n,),int)
	accumulator = np.zeros((n,),dtype=a.dtype)
	
	
	#dbg_array = []
	for i, delta in enumerate(deltas):
		_lgr.debug(f'calculating contributions from offset {delta} number {i}/{deltas.shape[0]}')
		offset_idx_array = (np.array(idx_array).T-delta).T
		values = index_boundary_func(a, offset_idx_array)
		_lgr.debug(f'{values=}')
		# Things that are not masked should contribute
		contrib = ~index_boundary_func(m, offset_idx_array)
		_lgr.debug(f'{contrib=}')
		values[~contrib] = 0
		accumulator += values
		contributions += contrib
		#dbg_array.append((delta, offset_idx_array, values, contrib))
	
	#a_old = np.array(a)
	_lgr.debug(f'{accumulator=}')
	_lgr.debug(f'{contributions=}')
	invalid = contributions == 0
	contributions[invalid] = 1
	accumulator[invalid] = const
	a[idx_array] = (accumulator/contributions)
	
	
	"""# Uncomment for debugging
	a_new = a
	m_plt = m
	if a.ndim > 2:
		st = tuple(slice(None) if i >= (a.ndim-2) else s//2 for i,s in enumerate(a.shape))
		_lgr.debug(f'{st=}')
		a_old = a_old[st]
		a_new = a_new[st]
		m_plt = m[st]
	import matplotlib.pyplot as plt
	import plot_helper as ph
	fig, ax = ph.figure_n_subplots(3)#+2*len(deltas))
	ax[0].imshow(a_old, origin='lower')
	ax[0].get_xaxis().set_visible(False)
	ax[0].get_yaxis().set_visible(False)
	
	ax[1].imshow(m_plt, origin='lower')
	ax[1].get_xaxis().set_visible(False)
	ax[1].get_yaxis().set_visible(False)
	
	ax[2].imshow(a_new, origin='lower')
	ax[2].get_xaxis().set_visible(False)
	ax[2].get_yaxis().set_visible(False)
	'''
	for i, (delta, offset_idx_array, values, contrib) in enumerate(dbg_array):
		print(f'{offset_idx_array} {values} {contrib}')
	
	for i, (delta, offset_idx_array, values, contrib) in enumerate(dbg_array):
		ax[3+2*i].imshow(contrib[None,:], interpolation='nearest', origin='lower')
		ax[3+2*i].set_title(f'{delta[0]}, {delta[1]}')
		ax[3+2*i].get_xaxis().set_visible(False)
		ax[3+2*i].get_yaxis().set_visible(False)
		ax[3+2*i+1].imshow(values[None,:], interpolation='nearest', vmin=np.min(a_new), vmax=np.max(a_new), origin='lower')
		ax[3+2*i+1].set_title(f'{delta[0]}, {delta[1]}')
		ax[3+2*i+1].get_xaxis().set_visible(False)
		ax[3+2*i+1].get_yaxis().set_visible(False)
		_lgr.debug(f'{contrib=}')
		_lgr.debug(f'{values=}')
	'''
	plt.show()
	"""
	
	

	return a



def interp(a : np.ndarray[S[N],T], m : np.ndarray[S[N],bool], fix_edges : bool = True, **kwargs) -> np.ndarray[S[N],T]:
	"""
	Replaces masked elements with linear interpolation of surrounding values in place.

	Arguments:
		a : np.ndarray
			A numpy array of pixel data
		m : np.ndarray
	"""
	from scipy.interpolate import griddata
	
	idxs = np.indices(a.shape)
	
	accumulator = griddata(np.moveaxis(idxs[:,~m],0,-1), a[~m], np.moveaxis(idxs[:,m],0,-1), **kwargs)
	a[m] = accumulator
	
	
	if np.any(np.isnan(accumulator)) and fix_edges:
		_lgr.debug('Fixing edges...')
		a = interp(a, np.isnan(a), fix_edges=False, **{**kwargs, 'method':'nearest'})
	
	if np.any(np.isnan(a)):
		raise RuntimeError('Could not interpolate masked elements, this may be due to NANs at the edges')
	
	return a



"""
Classes and routines that assist with plotting, specifically plotting the
progress of algorithms.
"""
from typing import Any, Callable
import dataclasses as dc
import copy
from pathlib import Path

import numpy as np
import matplotlib as mpl
import matplotlib.figure
import matplotlib.pyplot as plt

def output(
		show : bool | float = False, # Should we show the figure? If a number, then display figure for this number of seconds before moving onwards
		fname : str | Path | None = None, # If not None, save the figure to this file
		figure : mpl.figure.Figure | str | int | None = None, # The figure to operate on, default is current figure
		**kwargs # arguments to pass when saving figure, see plt.savefig(...)
	):
	"""
	Outputs the passed figure to a file `fname`, interactively `show` it, and close it afterwards
	"""
	if figure is not None:
		plt.figure(figure)
		
	if fname is not None:
		plt.savefig(fname, **kwargs)
	if type(show) is bool and show:
		plt.show()
	elif type(show) in (int,float):
		plt.pause(show)
		
	plt.close(figure)


def lowest_aspect_ratio_rectangle_of_at_least_area(x):
	sqrt_x = np.sqrt(x)
	b = int(sqrt_x)
	a = b
	while a*b < x:
		a += 1
	return(b,a)

def create_figure_with_subplots(nr, nc, nax=None ,size=6, squeeze=False, figure=None, flatten=True, fig_kwargs={}, sp_kwargs={}):
	"""
	Creates a figure and fills it with subplots
	
	# ARGUMENTS #
		nr
			<int> Number of rows
		nc
			<int> Number of columns
		nax
			<int> Number of axes to create
		size [2]
			<float> Size of figure to create x and y dimension, will be multipled by number of columns and rows
		fig_kwargs [dict]
			Figure keyword arguments. 'figsize' will overwrite passed values for 'size' if present.
		sp_kwargs [dict]
			Subplot keyword arguments. 'squeeze' will overwrite passed values if present. 

	# RETURNS #
		f
			Matplotlib figure created
		a [nax]
			List of axes contained in f
	"""
	# validate arguments
	if not hasattr(size, '__getitem__'):
		size = (size, size)
	if nax is None:
		nax = nr*nc

	# validate **kwargs
	if 'figsize' not in fig_kwargs:
		fig_kwargs['figsize'] = [nc*size[0], nr*size[1]]
	if 'squeeze' not in sp_kwargs:
		sp_kwargs['squeeze'] = squeeze
	
	if figure is None:
		f = plt.figure(**fig_kwargs)
	else:
		f = figure
	a = f.subplots(nr, nc, **sp_kwargs)
	
	for i, _ax in enumerate(a.flatten()):
		if i>=nax:
			_ax.remove()
	return(f, a.flatten() if flatten else a)

def figure_n_subplots(n, figure=None, fig_kwargs={}, sp_kwargs={}):
	"""
	Create a figure with `n` subplots arranged in a rectangle that has the lowest aspect ratio
	"""
	return(create_figure_with_subplots(
		*lowest_aspect_ratio_rectangle_of_at_least_area(n), 
		nax=n, size=6, squeeze=False, figure=figure, flatten=True,
		fig_kwargs=fig_kwargs,
		sp_kwargs=sp_kwargs)
	)


@dc.dataclass
class DiffClass:
	initial : Any = np.nan
	
	_stored : Any = dc.field(default=None, init=False, repr=False, compare=False)
	
	def set_initial(self, value):
		self.initial=value
		return self
	
	def __call__(self, value):
		prev = self._stored
		self._stored = copy.copy(value)
		return self.initial if prev is None else value - prev


def lim(data):
	a,b = np.nanmin(data), np.nanmax(data)
	return None if np.isnan(a) else a, None if np.isnan(b) else b
	

@dc.dataclass
class LimFixed:
	vmin : float | None = None
	vmax : float | None = None
	
	func : Callable[[Any],tuple[float,float]] = lim
	
	def __call__(self, data):
		a,b = self.func(data)
		if self.vmin is not None: a=self.vmin
		if self.vmax is not None: b=self.vmax
		return(a,b)


@dc.dataclass
class LimRememberExtremes:
	func : Callable[[Any],tuple[float,float]] = lim
	
	_vmin : float = None
	_vmax : float = None
	
	def __call__(self, data):
		a,b = self.func(data)
		if self._vmin is None or (a < self._vmin and a is not None): self._vmin = a
		if self._vmax is None or (b > self._vmax and b is not None): self._vmax = b
		return self._vmin, self._vmax

@dc.dataclass
class LimRememberNExtremes:
	n : int = 50
	func : Callable[[Any],tuple[float,float]] = lim
	
	_i : int = 0
	_vmin : tuple[int,float] = (0,None)
	_vmax : tuple[int,float] = (0,None)
	
	def __call__(self, data):
		#print(f'{data[-self.n:]=} {self._i=} {self.n=} {self._vmin=} {self._vmax=}')
		a,b = self.func(data[-self.n:])
		if a is not None and (self._vmin[0] < (self._i-self.n) or self._vmin[1] is None or a < self._vmin[1]): self._vmin = (self._i,a)
		if b is not None and (self._vmax[0] < (self._i-self.n) or self._vmax[1] is None or b > self._vmax[1]): self._vmax = (self._i,b)
		
		self._i += 1
		return self._vmin[1], self._vmax[1]

@dc.dataclass
class LimSymAroundCurrent:
	min_diff : float = 1.0
	func : Callable[[Any],tuple[float,float]] = lim
	
	def __call__(self, data):
		a, b = self.func(data)
		value = data[-1]
		farthest_from_value = np.nanmax(np.fabs([0 if a is None else a-value, 0  if b is None else b-value]))
		if farthest_from_value < self.min_diff:
			farthest_from_value = self.min_diff
		if np.isnan(farthest_from_value):
			return None,None
		return(None if a is None else -farthest_from_value + value, None if b is None else farthest_from_value + value)

@dc.dataclass
class LimSymAroundValue:
	value : float = 0
	pad : float = 0 # +ve numbers is a constant, -ve is a multiplier
	
	def __call__(self, data):
		farthest_from_value = np.nanmax(np.fabs(data-self.value))
		if np.isnan(farthest_from_value):
			return None,None
		if self.pad == 0: return(-farthest_from_value + self.value, farthest_from_value + self.value)
		elif self.pad > 0: return(-farthest_from_value + self.value - self.pad, farthest_from_value + self.value + self.pad)
		else: return((1+self.pad)*(-farthest_from_value + self.value), (1+self.pad)*(farthest_from_value + self.value))


@dc.dataclass
class LimAroundExtrema:
	factor : float = 0.1
	
	def __call__(self, data):
		dmin = np.nanmin(data)
		dmax = np.nanmax(data)
		return(dmin - np.fabs(dmin)*self.factor, dmax + np.fabs(dmax)*self.factor)	

def remove_axes_ticks_and_labels(ax, state=False):
	ax.xaxis.set_visible(state)
	ax.yaxis.set_visible(state)
	return

def remove_axis_all(ax, state=False):
	remove_axes_ticks_and_labels(ax, state)
	for spine in ax.spines.values():
		spine.set_visible(state)
	return


def flip_x_axis(ax):
	ax.set_xlim(ax.get_xlim()[::-1])
	return

def flip_y_axis(ax):
	ax.set_ylim(ax.get_ylim()[::-1])
	return

def get_legend_hls(*args):
	# gets handles and labels for each axis in *args
	hs, ls = [], []
	for ax in args:
		h, l = ax.get_legend_handles_labels()
		hs += h
		ls += l
	return(hs, ls)

def set_legend(*args, a_or_f=None, **kwargs):
	# sets the legend for the passed axes or figure "a_or_f" to the combination
	# of all handles and labels in each axes via "args"
	# **kwargs is passed to the ".legend()" method
	hs, ls = get_legend_hls(*args)
	if a_or_f is None:
		a_or_f = args[0]
	a_or_f.legend(hs, ls, **kwargs)
	return

def fig_add_patch(figure, *patches):
	figure.patches.extend(patches)

def fig_draw_bbox_of_artist(figure, *artists, color='red', linewidth=1, **kwargs):
	renderer = figure.canvas.get_renderer()
	for artist in artists:
		bbox = artist.get_window_extent(renderer)
		figure.lines.append(mpl.lines.Line2D(
			[bbox.x0, bbox.x0, bbox.x1, bbox.x1, bbox.x0], 
			[bbox.y0, bbox.y1, bbox.y1, bbox.y0, bbox.y0],
			linewidth=linewidth,
			color=color,
			**kwargs
		))



"""
Contains support routines for holding data about "strategies" (i.e., different ways of doing things) that perform "tasks" (e.g., functions, operations, methods).
"""

from typing import Callable, Any, Self
import dataclasses as dc


@dc.dataclass(init=True, repr=True, order=False, eq=False, slots=True)
class TaskStratInfo:
	"""
	Base class that describes a strategy to peform a task
	"""
	name : str = dc.field(default=dc.MISSING, init=True, repr=True, hash=False, compare=False)
	description : str = dc.field(default=dc.MISSING, init=True, repr=True, hash=False, compare=False)
	callable : Callable[[...],Any] = dc.field(default=dc.MISSING, init=True, repr=False, hash=False, compare=False)


@dc.dataclass(init=True, repr=True, order=False, eq=False, slots=True)
class TaskStratSet:
	"""
	A set of strategies that perform a task, having a collection makes it easy to choose one from another
	"""
	description : str = dc.field(default=dc.MISSING, init=True, repr=True, hash=False, compare=False)
	_items : list[TaskStratInfo] = dc.field(default_factory=list, init=False, repr=True, hash=False, compare=False)
	_name_idx_map : dict[str,int] = dc.field(default_factory=dict, init=False, repr=True, hash=False, compare=False)

	def _get_item(self, key : int | str) -> TaskStratInfo:
		if type(key) is int:
			return self._items[key]
		else:
			return self._items[self._name_idx_map[key]]
		raise RuntimeError('This should never happen as key should be an int or a string in the self._name_idx_map dict')
	
	def __getitem__(self, key : int | str) -> TaskStratInfo:
		return self._get_item(key)
	
	def _set_item(self, value : TaskStratInfo) -> Self:
		key : str = value.name
		idx = self._name_idx_map.get(key, len(self._items))
		self._name_idx_map[key] = idx
		self._items.append(value)
		return self
	
	def _del_item(self, key : int | str) -> Self:
		if type(key) is int:
			del self._items[key]
			
			to_del_key = None
			for k,v in self._name_idx_map.items():
				if v == key:
					to_del_key = k
					break
			if to_del_key is not None:
				del self._name_idx_map[key]
			else:
				raise RuntimeError('This should never happen as `key` should always be an int or a string in `self._name_idx_map`')
		else:
			idx = self._name_idx_map.get(key, None)
			if idx is not None:
				del self._items[idx]
			else:
				raise RuntimeError('This should never happen as `key` should be in the `self._name_idx_map` dictionary')
			del self._name_idx_map[key]
		return self
	
	def add(self, *args : list[TaskStratInfo,...]) -> Self:
		for value in args:
			if value.name not in self._name_idx_map:
				self._set_item(value)
			else:
				raise RuntimeError('Could not add value with name "{value.name}" as an item of that name already exists in the StrategySet')
		return self
	
	def remove(self, key) -> Self:
		if key in self._name_idx_map:
			self._del_item(key)
		# ignore key that is not present
		return self
	
	def get_callable(self, key : int | str) -> Callable[[...],Any]:
		return self.__getitem__(key).callable
	
	def get_description(self, key : int | str) -> Callable[[...],Any]:
		return self.__getitem__(key).description
	
	def get_name(self, key : int | str) -> Callable[[...],Any]:
		return self.__getitem__(key).name
	
	@property
	def names(self) -> tuple[str]:
		return tuple(item.name for item in self._items)
	
	@property
	def descriptions(self) -> tuple[str]:
		return tuple(item.description for item in self._items)
	
	@property
	def callables(self) -> tuple[Callable[[...],Any]]:
		return tuple(item.callable for item in self._items)
	
	def format_description(self, indent_level : int = 0, indent : str = '\t', with_default=True):
		lines = [indent*indent_level+self.description+f' (default : {self._items[0].name})' if with_default else '']
		for item in self._items:
			lines.append(indent*(indent_level+1)+item.name)
			lines.append(indent*(indent_level+2)+item.description)
		return '\n'.join(lines)
	
	def __call__(self, key, *args, **kwargs) -> Any:
		"""
		Call the strategy specified by 'key'
		"""
		return self.get_callable(key)(*args, **kwargs)
	



"""
Logging setup for the package
"""
import sys
import types
import logging

logging.basicConfig(
	format="%(asctime)s %(filename)s:%(lineno)d \"%(funcName)s\" %(levelname)s: %(message)s",
	datefmt="%Y-%m-%dT%H:%M:%S %z",
	force=True
)

excepthook_lgr = logging.getLogger('excepthook')
def log_excepthook(type, value, traceback):
	"""
	Override for `sys.excepthook`.
	"""
	
	# Go through the traceback and find out
	# where the exception was raised initially.
	tb = traceback
	while tb.tb_next is not None:
		tb = tb.tb_next
	
	# Use the original source of the exception as the source of the log.
	co = tb.tb_frame.f_code
	
	rcrd_fac = lambda msg: logging.LogRecord(excepthook_lgr.name, logging.ERROR, co.co_filename, tb.tb_lineno, msg, {}, None,func=co.co_qualname)
	
	rcrd = rcrd_fac(value)
	excepthook_lgr.handle(rcrd)
	#root_lgr.error(value, exc_info=(type, value, traceback), stack_info=True, stacklevel=5)
	if hasattr(value, '__notes__'):
		for note in value.__notes__:
			rcrd = rcrd_fac(f'NOTE: {note}')
			excepthook_lgr.handle(rcrd)
	sys.__excepthook__(type, value, traceback)

# Override except hook so we can log all uncaught exceptions
sys.excepthook = log_excepthook


def get_logger_at_level(name : str | types.ModuleType, level : str|int = logging.NOTSET) -> logging.Logger:
	""":
	Return the `name`d logger that reports `level` logs
	"""
	if type(name) is types.ModuleType:
		name = name.__name__
	
	_lgr = logging.getLogger(name)
	_lgr.setLevel(level)
	return _lgr

def set_logger_at_level(name : str | types.ModuleType, level : str|int = logging.NOTSET) -> None:
	"""
	Set the logger with `name` to report `level` logs
	"""
	if type(name) is types.ModuleType:
		name = name.__name__
	get_logger_at_level(name, level)
	return None



"""
Convert an image to FITS format
"""

import sys, os
from pathlib import Path
import dataclasses as dc
import argparse
import io
import re

from collections import namedtuple

from astropy.io import fits
import PIL
import PIL.Image
import PIL.features
import PIL.ExifTags


import numpy as np

import aopp_deconv_tool.astropy_helper as aph
import aopp_deconv_tool.astropy_helper.fits.header
from aopp_deconv_tool.fpath import FPath
import aopp_deconv_tool.arguments

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')


DataBundle = namedtuple('DataBundle', ('data', 'header'))

re_dashed_line = re.compile(r'\n-+\n') # lines just consisting of "-" characters
re_comma_space = re.compile(r',\s') # comma then a space

def get_supported_formats():
	ss = io.StringIO()
	PIL.features.pilinfo(out = ss, supported_formats=True)
	
	supported_extensions = []
	for chunk in re_dashed_line.split(ss.getvalue()):
		_lgr.debug(f'{chunk=}')
		extensions_are_supported = False
		possibly_supported_extensions = []
		
		for line in chunk.split('\n'):
			if line.startswith('Extensions: '):
				possibly_supported_extensions = re_comma_space.split(line[12:])
				_lgr.debug(f'{possibly_supported_extensions=}')
			if line.startswith('Features: '):
				if 'open' in line[10:]:
					extensions_are_supported = True
		if extensions_are_supported:
			supported_extensions.extend(possibly_supported_extensions)
			
	return supported_extensions


def read_exif(image, header={}, exif_tag_reader=lambda k, v: (str(k), v)):
	exif = image.getexif()
	_lgr.debug(f'{exif=}')
	for k,v in exif.items():
		tag, value = exif_tag_reader(k,v)
		header[tag] = value
		_lgr.debug(f'{k} : {v}')
	_lgr.debug(f'DONE')
	return header

def read_image_into_numpy_array(fpath : Path) -> DataBundle:
	match fpath.suffix:
		case '.tif' | '.tiff':
			exif_tag_reader = lambda k, v: (PIL.TiffTags.lookup(k).name, v)
		case _:
			exif_tag_reader=lambda k, v: (str(k), v)
	
	
	data = None	
	header = {}
	with PIL.Image.open(fpath) as image:
		header = read_exif(image, header=header, exif_tag_reader=exif_tag_reader)
		image = image.convert(mode='F')
		data = np.array(image).astype(np.float64)
		
	return DataBundle(data, header)


def save_as_fits(output_path : str | Path, primary_data_bundle : DataBundle, **kwargs : dict[str, DataBundle]):

	hdr = fits.Header()
	hdr.update(aph.fits.header.DictReader(primary_data_bundle.header))
	hdu_primary = fits.PrimaryHDU(
		header = hdr,
		data = primary_data_bundle.data
	)
	
	hdu_list = [hdu_primary]
	
	for k,v in kwargs.items():
		new_hdr = fits.Header()
		new_hdr.update(aph.fits.header.DictReader(v.header))
		hdu_list.append(fits.ImageHDU(header=new_hdr, name=k, data=v.data))
	
	
	hdul = fits.HDUList(hdu_list)
	
	hdul.writeto(output_path, overwrite=True)
	_lgr.info(f'Converted file written to "{output_path}"')


def parse_args(argv):
	parser = argparse.ArgumentParser(
		description=__doc__,
	)
	
	SUPPORTED_FORMATS = get_supported_formats()
	_lgr.debug(f'{SUPPORTED_FORMATS=}')
	
	parser.add_argument('image_path', type=Path, help=f'Path to the image to convert, can be one of {" ".join(SUPPORTED_FORMATS)}')
	#parser.add_argument('-o', '--output_path', type=Path, default=None, help='Path to save fits conversion to, if not supplied will replace original file extension with ".fits"')
	parser.add_argument(
		'-o', 
		'--output_path', 
		type=FPath,
		metavar='str',
		default='{parent}/{stem}.fits',
		help = '\n    '.join((
			f'Output fits file path, supports keyword substitution using parts of `image_path` path where:',
			'{parent}: containing folder',
			'{stem}  : filename (not including final extension)',
			'{suffix}: final extension (everything after the last ".")',
			'\b'
		))
	)
	
	args = parser.parse_args(argv)
	
	if args.image_path.suffix not in SUPPORTED_FORMATS:
		parser.print_help()
		_lgr.error(f'Unsupported image format {args.image_path.suffix}')
		sys.exit()
	
	#if args.output_path is None:
	#	args.output_path = Path(str(args.image_path.stem) + '.fits')
	other_file_path = Path(args.fits_spec.path)
	args.output_path = args.output_path.with_fields(
		tag=DEFAULT_OUTPUT_TAG, 
		parent=other_file_path.parent, 
		stem=other_file_path.stem, 
		suffix=other_file_path.suffix
	)
	
	for k,v in vars(args).items():
		_lgr.debug(f'{k} = {v}')
	
	return args



def go(
		image_path,
		output_path=None
	):
	"""
	Thin wrapper around `run()` to accept string inputs.
	As long as the names of the arguments to this function 
	are the same as the names expected from the command line
	we can do this programatically
	"""
	# This must be first so we only grab the arguments to the function
	fargs = dict(locals().items())
	arglist = aopp_deconv_tool.arguments.construct_arglist_from_locals(fargs, n_positional_args=1)
	
	exec_with_args(arglist)
	return

def exec_with_args(argv):
	args = parse_args(argv)
	
	
	data_bundle = read_image_into_numpy_array(args.image_path)
	save_as_fits(args.output_path, data_bundle)

if __name__=='__main__':
	exec_with_args(sys.argv[1:])


"""
Quick tool for spectrally rebinning a FITS file, convolves a triangular 
function of `bin_width` with the input data and samples it every `bin_step`.

NOTE: Supersampling is not done in any special way, so if that is desired, 
set `bin_width` to the distance between two datapoints and `bin_step` to the 
supersampling length. This will not preserve the sum, but it should work when
interpolation is required.
"""

import sys
from pathlib import Path
import dataclasses as dc
from typing import Literal, Callable, Sequence



import numpy as np
import scipy as sp
import scipy.signal

from astropy.io import fits
from astropy import units as u

import aopp_deconv_tool.astropy_helper as aph
import aopp_deconv_tool.astropy_helper.fits.specifier
import aopp_deconv_tool.astropy_helper.fits.header
import aopp_deconv_tool.numpy_helper as nph
import aopp_deconv_tool.numpy_helper.axes
import aopp_deconv_tool.numpy_helper.slice
import aopp_deconv_tool.numpy_helper.array.index
import aopp_deconv_tool.arguments

import aopp_deconv_tool.numpy_helper.array.grid
from aopp_deconv_tool.fpath import FPath

import matplotlib.pyplot as plt

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')



DEBUG_PIXEL = (161,168)


@dc.dataclass
class ResponseFunction:
	total_sum : float | None = None
	
	def as_array(self, pos_array : np.ndarray, trim : bool = True) -> np.ndarray:
		NotImplemented
	
	def get_pos_array(self, step:float) -> np.ndarray:
		NotImplemented
	

@dc.dataclass
class SquareResponseFunction (ResponseFunction):
	full_width_half_maximum : float = dc.field(default=2E-9, init=True)
	
	def as_array(self, pos_array : np.ndarray, trim : bool = True) -> np.ndarray:
		response_array = np.interp(pos_array, (0,self.full_width_half_maximum), (1,1), left=0, right=0)
		if self.total_sum is not None:
			response_array *= self.total_sum/np.sum(response_array)
		if trim:
			min_idx = np.max(np.argwhere(pos_array<=0))
			max_idx = np.min(np.argwhere(pos_array>=self.full_width_half_maximum))+1
			response_array = response_array[min_idx:max_idx]
		return response_array
	
	def get_pos_array(self, step : float) -> np.ndarray:
		return np.linspace(0, self.full_width_half_maximum, int(np.ceil(self.full_width_half_maximum/step)))
	

@dc.dataclass
class TriangularResponseFunction (ResponseFunction):
	full_width_half_maximum : float = dc.field(default=2E-9, init=True)
	
	def as_array(self, pos_array : np.ndarray, trim=True) -> np.ndarray:
		response_array = np.interp(pos_array, (0,self.full_width_half_maximum,2*self.full_width_half_maximum), (0,1,0), left=0, right=0)
		_lgr.debug(f'{pos_array=}')
		_lgr.debug(f'{response_array=}')
		
		if self.total_sum is not None:
			response_array *= self.total_sum/np.sum(response_array)
		
		if trim:
			min_idx = np.max(np.argwhere(pos_array<=0))
			max_idx = np.min(np.argwhere(pos_array>=2*self.full_width_half_maximum))+1
			response_array = response_array[min_idx:max_idx]
		
		return response_array
	
	def get_pos_array(self, step : float) -> np.ndarray:
		return np.linspace(0, 2*self.full_width_half_maximum, int(np.ceil(2*self.full_width_half_maximum/step)))
	

named_spectral_binning_parameters = dict(
	spex = dict(
		bin_step = 1E-9,
		bin_width = 2E-9,
		response_function_class = TriangularResponseFunction
	)
)

def indices_const_boundary(
		indices : tuple[Sequence[int]], 
		shape : tuple[int]
	) -> tuple[int]:
	"""
	Give a tuple with N entries `indices` like ((i_00,i_01,i_02,...,i_0M_0),(i_10,i_11,...,i_1M_1),...,(...,i_NM_N)), where N is the number of entries in `shape`
	which is like (s_0, s_1, ..., s_N), and M_0, M_1, ..., M_N are the number of indices present for each axis.
	
	Alter the indices so if i_nm < 0, i_nm=0. And if i_nm >= s_n, i_nm=s_n-1. I.e., indices "to the left" are set to the first entry in the axis, and
	indices "to the right" are set to the last entry in the axis.
	"""
	out_indices = []
	for j, (idxs, s) in enumerate(zip(indices, shape)):
		jdx = []
		for i in idxs:
			if i <0:
				jdx.append(0)
			elif i >= s:
				jdx.append(s-1)
			else:
				jdx.append(i)
		
		out_indices.append(tuple(jdx))
	return tuple(out_indices)

def array_const_boundary(a : np.ndarray, indices: tuple[Sequence[int]]) -> np.ndarray:
	"""
	Return the values on an arry such that any indices that are "to the left" return the first entry in the axis, and any that are "to the right" return
	the last entry in the axis.
	"""
	return a[indices_const_boundary(indices, a.shape)]


def overlap_add_convolve(
		data : np.ndarray[[...,'M',...]], 
		response_1d : np.ndarray[['N']], 
		axis : int, 
		mode : Literal['same'] | Literal['full'] | Literal['valid'] = 'same'
	) -> np.ndarray[[...,'L',...]]:
	"""
	Performs convolution over `axis` of `data` by the overlap-and-add method with `response_1d`. Is generally faster for large `data` and small `response_1d`.
	The shape of the output array depends on the `mode`:
	
	mode    | output shape
	--------|-----------------
	'same'  | [...,'M',...]
	'full'  | [...,'M+2N',...]
	'valid' | [...,'M-2N',...]
	"""
	data = np.moveaxis(data, 0, axis)
	
	n = response_1d.shape[0]
	full = np.zeros((data.shape[0]+2*n, *data.shape[1:]))
	
	
	
	for i in range(-n, data.shape[0]+n):
		subset_of_data = array_const_boundary(data, (np.arange(i,i+n,dtype=int),))
		full[i+n] = np.sum((subset_of_data.T * response_1d[:subset_of_data.shape[0]]).T, axis=0)
	
	
	dn = 1
	
	"""
	# DEBUGGING
	# Plot data and convolved data on the same axis to work out offset needed
	full_embed_data = np.zeros_like(full)
	full_embed_data[n//2+dn:-2*n+n//2+dn] = data
	plt.figure()
	plt.plot(full_embed_data[:,*DEBUG_PIXEL])
	plt.plot(full[:,*DEBUG_PIXEL])
	plt.show()
	"""
	
	match mode:
		case 'same':
			return np.moveaxis(full[n//2+dn:-2*n+n//2+dn], 0, axis)
		case 'full':
			return np.moveaxis(full, 0, axis)
		case 'valid':
			return np.moveaxis(full[n+dn:-2*n+dn], 0, axis)
		case _:
			raise RuntimeError(f'Unknown mode "{mode}"')

def lin_interp(
		new_points : np.ndarray[['N']], 
		old_points : np.ndarray[['M']], # assume this is sorted in ascending order
		data : np.ndarray[[...,'M',...]], 
		axis : int, # should be axis of "M" in `data`,
		boundary_conditions : Literal['constant'] | Literal['reflect'] | Literal['periodic'] | Literal['extrapolate'] = 'constant'
	) -> np.ndarray:
	"""
	Given `data` defined at `old_points` along `axis`, linearly interpolate at `new_points` subject to `boundary_conditions`.
	Works on N-dimensional arrays, but only over one axis at a time.
	"""
	
	data = np.moveaxis(data, axis, -1)
	_lgr.debug(f'Moved data axis')
	
	y = np.full((*data.shape[:-1], new_points.size), fill_value=np.nan)
	_lgr.debug(f'Output array allocated')
	
	new_points = np.array(new_points) # copy array soe we can process the points
	match boundary_conditions:
		case 'constant':
			gt = new_points >= old_points[-1]
			lt = new_points < old_points[0]
			new_points[gt] = old_points[-1]
			new_points[lt] = old_points[0]
		case 'reflect':
			gt = new_points >= old_points[-1]
			lt = new_points < old_points[0]
			while gt.size!=0 or lt.size !=0:
				new_points[lt] = 2*old_points[0] - new_points[lt]
				new_points[gt] = 2*old_points[1] - new_points[gt]
				gt = new_points >= old_points[-1]
				lt = new_points < old_points[0]
		case 'periodic':
			gt = new_points >= old_points[-1]
			lt = new_points < old_points[0]
			while gt.size!=0 or lt.size !=0:
				new_points[lt] = old_points[0] - new_points[lt] + old_points[1]
				new_points[gt] -=  old_points[-1] - old_points[0]
				gt = new_points >= old_points[-1]
				lt = new_points < old_points[0]
		case 'extrapolate':
			pass
		case _:
			raise RuntimeError(f'Unknown boundary conditions "{boundary_conditions}"')
	_lgr.debug(f'Calculated boundary conditions')
	
	
	_lgr.debug(f'{data.shape=} {old_points.shape=}')
	
	
	new_idxs = np.zeros_like(new_points, dtype=int)	
	for i in range(1, old_points.size-1):
		new_idxs += (old_points[i] <= new_points).astype(int)
		
	_lgr.debug(f'Calculated indices of new points')
	
	if (data.size < 1E6):
		# Do things the fast way if we can
		A = np.full((*data.shape[:-1], old_points.size-1, 2), fill_value=np.nan, dtype=float)
		
		A[...,:,0] = (data[...,1:] - data[...,:-1])/(old_points[1:] - old_points[:-1]) # gradient
		A[...,:,1] = data[...,:-1] - A[...,0]*old_points[:-1] # constant
		
		y[...] = np.sum(A[...,new_idxs,:]*np.stack((new_points, np.ones_like(new_points)), axis=-1), axis=-1)
		
		"""
		# DEBUGGING
		plt.figure()
		plt.title('linear interpolation parameters')
		ax = plt.gca()
		ax.plot(A[*DEBUG_PIXEL,:,0], color='tab:blue', alpha=0.5, label='gradient')
		ax.set_ylabel('gradient')
		ax2 = ax.twinx()
		ax2.plot(A[*DEBUG_PIXEL,:,1], color='tab:orange', alpha=0.5, label='constant')
		ax2.set_ylabel('constant')
		ax_h = ax.get_legend_handles_labels()
		ax2_h = ax2.get_legend_handles_labels()
		plt.legend(ax_h[0]+ax2_h[0], ax_h[1]+ax2_h[1])
		plt.show()
		plt.close('all')
		"""
		return np.moveaxis(y, -1, axis)
	
	# Otherwise, do things piecewise so we don't run out of memory
	
	for i, j in enumerate(new_idxs):
		if 1 <= j < old_points.size:
			gradient = (data[...,j] - data[...,j-1])/(old_points[j] - old_points[j-1])
		else:
			gradient = np.zeros(data.shape[:-1])

		if 0 <= j <= old_points.size:
			constant = data[...,j] - gradient*old_points[j]
		else:
			constant = 0
			
		y[...,i] = gradient*new_points[i] + constant
		
	return np.moveaxis(y,-1,axis)
	

def plot_rebin_response_function(data_hdu, axis, spectral_unit_in_meters, new_spec_ax, new_data):
	_lgr.debug(f'{new_spec_ax.shape=} {new_data[:,*DEBUG_PIXEL].shape=}')
	input_axis = aph.fits.header.get_world_coords_of_axis(
		data_hdu.header, 
		axis, 
		wcs_unit_to_return_value_conversion_factor=spectral_unit_in_meters
	)
	plt.plot(
		input_axis, 
		data_hdu.data[:,*DEBUG_PIXEL],
		color='tab:blue',
		label='original'
	)
	plt.plot(
		new_spec_ax, 
		new_data[:,*DEBUG_PIXEL],
		color='tab:orange',
		label='rebinned'
	)
	plt.plot(
		input_axis, 
		data_hdu.data[:,*DEBUG_PIXEL] - np.interp(input_axis, new_spec_ax, new_data[:,*DEBUG_PIXEL]),
		color='tab:green',
		label='original - rebinned'
	)
	plt.legend()
	plt.show()


def rebin_hdu_over_axis_with_response_function(
		data_hdu,
		axis,
		response_function : ResponseFunction,
		bin_start : float | None = None,
		bin_step : float = 1E-9,
		axis_unit_conversion_factors : tuple[float,...] = (1,),
		plot : bool = True
	) -> tuple[np.ndarray, np.ndarray]:
	
	_lgr.debug(f'{axis_unit_conversion_factors=}')
	ax_values = aph.fits.header.get_world_coords_of_axis(data_hdu.header, axis, wcs_unit_to_return_value_conversion_factor=axis_unit_conversion_factors)
	_lgr.debug(f'{data_hdu.data.shape=} {ax_values.shape=}')
	
	bin_start = ax_values[0] if bin_start is None else bin_start
	
	current_step = ax_values[1]-ax_values[0]
	if current_step > bin_step:
		_lgr.warning(f'Current rebinning does not work well for reducing step size, {current_step=} {bin_step=}. Using current data instead')
		return ax_values, data_hdu.data
	
	response_array = response_function.as_array(ax_values-ax_values[0], trim=True)
	_lgr.debug(f'{response_array=}')
	
	#plt.plot(response_array); plt.show()
	
	smoothed = overlap_add_convolve(
		data_hdu.data, 
		response_array,
		axis=axis,
		mode='same'
	)
	
	n_new_points = (ax_values[-1]-bin_start)/bin_step + 1
	_lgr.debug(f'{n_new_points=}')
	new_points = np.linspace(bin_start, ax_values[-1], int(np.floor(n_new_points)))
	smoothed = lin_interp(new_points, ax_values, smoothed, axis=axis, boundary_conditions='constant')
	
	if plot:
		plot_rebin_response_function(data_hdu, axis, axis_unit_conversion_factors, new_points, smoothed)
	
	
	return new_points, smoothed


def run(
		fits_spec : aph.fits.specifier.FitsSpecifier, 
		output_path : Path | str, 
		bin_step : float = 1E-9, # Assume SI units
		bin_width : float = 2E-9, # Assume SI units
		operation : Literal['sum'] | Literal['mean'] | Literal['mean_err'] = 'mean',
		spectral_unit_in_meters : float = 1,
		response_function_class : Literal[SquareResponseFunction] | Literal[TriangularResponseFunction] = TriangularResponseFunction,
		output_unit : u.Unit | None = None,
		plot : bool = False,
	) -> tuple[np.ndarray, np.ndarray]:

	original_data_type=None

	new_data = None
	with fits.open(Path(fits_spec.path)) as data_hdul:
	
		#_lgr.debug(f'{fits_spec.ext=}')
		#raise RuntimeError(f'DEBUGGING')
	
		data_hdu = data_hdul[fits_spec.ext]
		
	
		axes_ordering =  aph.fits.header.get_axes_ordering(data_hdu.header, fits_spec.axes['SPECTRAL'])
		axis_unit = u.Unit(aph.fits.header.get_axes_unit_string(data_hdu.header, (axes_ordering[0].fits,))[0], format='fits')
		if output_unit is None:
			output_unit = axis_unit
		axis = axes_ordering[0].numpy
		original_data_type = data_hdu.data.dtype
				
		axis_si_unit = None
		match axis_unit.physical_type:
			case 'length':
				axis_si_unit = u.meter
			case 'frequency':
				axis_si_unit = u.Hz
			case _:
				raise RuntimeError(f'Unsupported physical type {axis_unit.physical_type} for unit {axis_unit}')
	
		#new_spec_bins, new_data = rebin_hdu_over_axis(data_hdu, axis, bin_step, bin_width, operation, plot=False)
		
		postprocess_data_mutator = lambda x: x
		
		response_function_sum = None
		match operation:
			case 'sum':
				pass # no need to alter anything
			case 'mean':
				response_function_sum = 1
			case 'mean_err':
				data_hdu.data = data_hdu.data**2 # square the data
				response_function_sum = 1
				postprocess_data_mutator = lambda x: np.sqrt(x) # square-root it after rebinning
			case _:
				raise RuntimeError(f'Unknown binning operation "{operation}"')
		
		new_spec_ax, new_data = rebin_hdu_over_axis_with_response_function(
			data_hdu, 
			axis, 
			response_function_class(response_function_sum, bin_width),
			bin_start=None, 
			bin_step=bin_step, 
			axis_unit_conversion_factors=1, # astropy always responds with SI units
			plot=plot
		)
		
		new_data = postprocess_data_mutator(new_data)
		
		
		hdr = data_hdu.header
		axis_fits = axes_ordering[0].fits
		param_dict = {
			'original_file' : Path(fits_spec.path).name, # record the file we used
			'bin_axis' : axis_fits,
			'bin_step' : bin_step,
			'bin_width' : bin_width,
			'bin_operation' : operation
		}
		
		hdr.update(aph.fits.header.DictReader(
			param_dict,
			prefix='spectral_rebin',
			pkey_count_start=aph.fits.header.DictReader.find_max_pkey_n(hdr)
		))
		
		#unit_factor = (1*output_unit / u.meter).value
		#unit_factor = (1*u.meter).to(axis_unit, equivalencies=u.spectral()).to(output_unit, equivalencies=u.spectral()).value
		unit_factor = (1*axis_si_unit).to(output_unit, equivalencies=u.spectral()).value
		aph.fits.header.set_axes_transform(hdr, 
			axis_fits, 
			u.format.Fits.to_string(output_unit), 
			new_spec_ax[0] * unit_factor,
			bin_step * unit_factor,
			new_spec_ax.shape[0],
			1
		)

	
	# Save the products to a FITS file
	hdu_rebinned = fits.PrimaryHDU(
		header = hdr,
		data = new_data.astype(original_data_type)
	)
	hdul_output = fits.HDUList([
		hdu_rebinned,
	])
	hdul_output.writeto(output_path, overwrite=True)
	_lgr.info(f'Written processed file to "{output_path}"')




def parse_args(argv):
	import os
	import aopp_deconv_tool.text
	import argparse
	
	DEFAULT_OUTPUT_TAG = '_rebin'
	DESIRED_FITS_AXES = ['SPECTRAL']
	OUTPUT_COLUMNS=80
	try:
		OUTPUT_COLUMNS = os.get_terminal_size().columns - 30
	except Exception:
		pass
	
	FITS_SPECIFIER_HELP = aopp_deconv_tool.text.wrap(
		aph.fits.specifier.get_help(DESIRED_FITS_AXES).replace('\t', '    '),
		OUTPUT_COLUMNS
	)
	
	parser = argparse.ArgumentParser(
		description=__doc__, 
		formatter_class=argparse.RawTextHelpFormatter,
		epilog=FITS_SPECIFIER_HELP
	)
	
	parser.add_argument(
		'fits_spec',
		help = '\n'.join((
			f'FITS Specifier of the data to operate upon . See the end of the help message for more information',
			f'required axes: {", ".join(DESIRED_FITS_AXES)}',
		)),
		type=str,
		metavar='FITS Specifier',
	)
	#parser.add_argument('-o', '--output_path', help=f'Output fits file path. By default is same as the `fits_spec` path with "{DEFAULT_OUTPUT_TAG}" appended to the filename')
	parser.add_argument(
		'-o', 
		'--output_path', 
		type=FPath,
		metavar='str',
		default='{parent}/{stem}{tag}{suffix}',
		help = '\n    '.join((
			f'Output fits file path, supports keyword substitution using parts of `fits_spec` path where:',
			'{parent}: containing folder',
			'{stem}  : filename (not including final extension)',
			f'{{tag}}   : script specific tag, "{DEFAULT_OUTPUT_TAG}" in this case',
			'{suffix}: final extension (everything after the last ".")',
			'\b'
		))
	)
	
	parser.add_argument('--spectral_unit_in_meters', type=float, default=None, help='The conversion factor between the spectral unit and meters. Only required when the unit cannot be automatically determined, or there is a mismatch between unit and data. Any automatically found unit information will be overwritten.')
	parser.add_argument('--rebin_operation', choices=['sum', 'mean', 'mean_err'], default='mean', help='Operation to perform when binning.')
	parser.add_argument('--output_unit', type=str, default=None, help='Units to use for output datacube. If not specified will use the same units as the input datacube.')
	
	rebin_group = parser.add_mutually_exclusive_group(required=False)
	rebin_group.add_argument('--rebin_preset', choices=list(named_spectral_binning_parameters.keys()), default='spex', help='Rebin according to the spectral resolution of the preset')
	rebin_group.add_argument('--rebin_params', nargs=2, type=float, metavar='float', help='bin_step and bin_width for rebinning operation (meters)')
	
	args = parser.parse_args(argv)
	
	args.fits_spec = aph.fits.specifier.parse(args.fits_spec, DESIRED_FITS_AXES)
	
	if args.output_unit is not None:
		try:
			args.output_unit = u.Unit(args.output_unit, format='fits')
		except ValueError as e:
			try:
				desired_unit = u.Unit(args.output_unit)
			except Exception as e:
				raise e
			else:
				equiv_phys_types = {}
				fits_compat_units = u.format.Fits._generate_unit_names()[0]
				compat_units = desired_unit.compose(
					#equivalencies=u.spectral(), 
					units=fits_compat_units.values(), 
					include_prefix_units=False
				)
				e.add_note(f'Acceptible equivalent FITS units for "{args.output_unit}":\n\t'+'\n\t'.join([str(x) for x in compat_units]))
				raise e
	
	if args.rebin_preset is not None:
		for k,v in named_spectral_binning_parameters[args.rebin_preset].items():
			setattr(args, k, v)
	if args.rebin_params is not None:
		setattr(args, 'bin_step', args.rebin_params[0])
		setattr(args, 'bin_width', args.rebin_params[1])
		setattr(args, 'response_function_class', TriangularResponseFunction)
	
	#if args.output_path is None:
	#	args.output_path =  (Path(args.fits_spec.path).parent / (str(Path(args.fits_spec.path).stem)+DEFAULT_OUTPUT_TAG+str(Path(args.fits_spec.path).suffix)))
	other_file_path = Path(args.fits_spec.path)
	args.output_path = args.output_path.with_fields(
		tag=DEFAULT_OUTPUT_TAG, 
		parent=other_file_path.parent, 
		stem=other_file_path.stem, 
		suffix=other_file_path.suffix
	)
	
	
	print('INPUT PARAMETERS')
	for k,v in vars(args).items():
		print(f'    {k} : {v}')
	print('END')
	
	return args


def go(
		fits_spec,
		output_path=None,
		spectral_unit_in_meters=None,
		rebin_operation=None,
		output_unit=None,
		rebin_preset=None,
		rebin_params=None
	):
	"""
	Thin wrapper around `run()` to accept string inputs.
	As long as the names of the arguments to this function 
	are the same as the names expected from the command line
	we can do this programatically
	"""
	# This must be first so we only grab the arguments to the function
	fargs = dict(locals().items())
	arglist = aopp_deconv_tool.arguments.construct_arglist_from_locals(fargs, n_positional_args=1)
	
	exec_with_args(arglist)
	return

def exec_with_args(argv):
	args = parse_args(argv)
	
	run(
		args.fits_spec, 
		bin_step=args.bin_step, 
		bin_width=args.bin_width, 
		operation=args.rebin_operation, 
		output_path=args.output_path, 
		spectral_unit_in_meters= 1 if args.spectral_unit_in_meters is None else args.spectral_unit_in_meters,
		response_function_class = args.response_function_class,
		output_unit = args.output_unit,
	)

if __name__ == '__main__':
	exec_with_args(sys.argv[1:])
	



from typing import Any
import dataclasses as dc

import numpy as np
import scipy as sp
import scipy.signal

from .base import Base

from aopp_deconv_tool.data_structures.pointer import Pointer
import aopp_deconv_tool.image_processing as im_proc
import aopp_deconv_tool.image_processing.otsu_thresholding
import aopp_deconv_tool.mfunc as mfunc

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')


@dc.dataclass(slots=True,repr=False)
class CleanModified(Base):
	"""
	A modified verison of the CLEAN algorithm, designed to account for extended emission.
	
	# ARGUMENTS #
		n_iter : int = 1000
			Number of iterations to perform
		loop_gain : float = 0.02
			Fraction of emission that could be accounted for by a PSF added to components with each iteration.
			Higher values are faster but unstable
		threshold : float = 0.3
			Fraction of maximum brightness of residual above which pixels will be included in CLEAN step, if 
			negative will use the "maximum fractional difference otsu threshold". 0.3 is a  good default value, 
			if stippling becomes an issue, reduce or set to a negative value. Lower positive numbers will 
			require more iterations, but give a more "accurate" result.
		n_positive_iter : int = 0
			Number of iterations to do that only "adds" emission, before switching to "adding and subtracting" 
			emission.
		noise_std : float = 1E-1
			Estimate of the deviation of the noise present in the observation
		rms_frac_threshold : float = 1E-3
			Fraction of original RMS of residual at which iteration is stopped, lower values continue iteration 
			for longer.
		fabs_frac_threshold : float = 1E-3
			Fraction of original Absolute Brightest Pixel of residual at which iteration is stopped, lower 
			values continue iteration for longer.
		max_stat_increase : float = np.inf
			Maximum fractional increase of a statistic before terminating
		min_frac_stat_delta : float = 1E-3
			Minimum fractional standard deviation of statistics before assuming no progress is being made and 
			terminating iteration.
		give_best_result : bool = True
			If True, will return the best (measured by statistics) result instead of final result.
	
	# RETURNS #
		A "CleanModified" instance. Run the model using the __call__(...) method.
		e.g.
		```
		from aopp_deconv_tool.algorithm.deconv.clean_modified import CleanModified
		
		deconvolver = CleanModified()
		...
		deconv_components, deconv_residual, deconv_iters = deconvolver(processed_obs, normed_psf)
		
		```
		See `aopp_deconv_tool.deconvolve` for a full example.
	"""
	n_iter 				: int 	= dc.field(default=1000, 	metadata={
		'description':'Number of iterations', 
		'domain':(0,np.inf),
	})
	loop_gain 			: float = dc.field(default=0.02, 	metadata={
		'description':'Fraction of emission that could be accounted for by a PSF added to components each iteration. Higher values are faster, but unstable.',
		'domain' : (0,1),
	})
	threshold 			: float = dc.field(default=0.3, 	metadata={
		'description':'Fraction of maximum brightness of residual above which pixels will be included in CLEAN step, if negative will use the maximum fractional difference otsu threshold. 0.3 is a  good default value, if stippling becomes an issue, reduce or set to a negative value. Lower positive numbers will require more iterations, but give a more "accurate" result.',
		'domain' : (0,1),
	})
	n_positive_iter 	: int 	= dc.field(default=0, 		metadata={
		'description':'Number of iterations to do that only "adds" emission, before switching to "adding and subtracting" emission',
		'domain' : (0,np.inf),
	})
	noise_std 			: float = dc.field(default=1E-1, 	metadata={
		'description':'Estimate of the deviation of the noise present in the observation, at the moment only used for calculating generalised least squares statistic',
		'domain' : (0, np.inf),
	})
	rms_frac_threshold 	: float = dc.field(default=1E-2, 	metadata={
		'description':'Fraction of original RMS of residual at which iteration is stopped, lower values continue iteration for longer.',
		'domain' : (0,1),
	})
	fabs_frac_threshold : float = dc.field(default=1E-2, 	metadata={
		'description':'Fraction of original Absolute Brightest Pixel of residual at which iteration is stopped, lower values continue iteration for longer.',
		'domain' : (0,1),
	})
	max_stat_increase	: float = dc.field(default=np.inf, 	metadata={
		'description':'Maximum fractional increase of a statistic before terminating',
		'domain' : (0, np.inf),
	})
	min_frac_stat_delta	: float = dc.field(default=1E-3, 	metadata={
		'description':'Minimum fractional standard deviation of statistics before assuming no progress is being made and terminating iteration',
		'domain' : (0,1),
	})
	give_best_result	: bool  = dc.field(default=False, 	metadata={
		'description':'If True, will return the best (measured by statistics) result instead of final result.',
		'domain' : {False, True},
	})
	clean_beam_sigma	: float = dc.field(default=0, 		metadata={
		'description':'If not zero, will convolve the components with a gaussian with this sigma, useful for regularising the result',
		'domain' : (0, np.inf),
	})
	
	# private attributes
	_obs                         : np.ndarray     = dc.field(init=False, repr=False, hash=False, compare=False)
	_psf                         : np.ndarray     = dc.field(init=False, repr=False, hash=False, compare=False)
	_residual_copy               : np.ndarray     = dc.field(init=False, repr=False, hash=False, compare=False)
	_iter_stat_record            : Any            = dc.field(init=False, repr=False, hash=False, compare=False)
	_iter_stat_names             : tuple[str,...] = dc.field(init=False, repr=False, hash=False, compare=False)
	_selected_map                : np.ndarray     = dc.field(init=False, repr=False, hash=False, compare=False)
	_accumulator                 : np.ndarray     = dc.field(init=False, repr=False, hash=False, compare=False)
	_current_convolved           : np.ndarray     = dc.field(init=False, repr=False, hash=False, compare=False)
	_current_cleaned             : np.ndarray     = dc.field(init=False, repr=False, hash=False, compare=False)
	_selected_px                 : np.ndarray     = dc.field(init=False, repr=False, hash=False, compare=False)
	_tmp_r                       : np.ndarray     = dc.field(init=False, repr=False, hash=False, compare=False)
	_px_choice_img_ptr           : Pointer        = dc.field(init=False, repr=False, hash=False, compare=False)
	_loop_gain_correction_factor : Any            = dc.field(init=False, repr=False, hash=False, compare=False)
	_flux_correction_factor      : Any            = dc.field(init=False, repr=False, hash=False, compare=False)
	_choose_update_pixels        : Any            = dc.field(init=False, repr=False, hash=False, compare=False)
	_pixel_threshold             : Any            = dc.field(init=False, repr=False, hash=False, compare=False)
	_get_pixel_threshold         : Any            = dc.field(init=False, repr=False, hash=False, compare=False)
	_fabs_threshold              : float          = dc.field(init=False, repr=False, hash=False, compare=False)
	_rms_threshold               : float          = dc.field(init=False, repr=False, hash=False, compare=False)
	_components_best             : np.ndarray     = dc.field(init=False, repr=False, hash=False, compare=False)
	_stats_best                  : np.ndarray     = dc.field(init=False, repr=False, hash=False, compare=False)
	_i_best                      : int            = dc.field(init=False, repr=False, hash=False, compare=False)
	_stats_delta                 : np.ndarray     = dc.field(init=False, repr=False, hash=False, compare=False)

	def get_components(self) -> np.ndarray:
		#return(self._components)
		if self.clean_beam_sigma == 0:
			return(self._components_best if self.give_best_result else self._components)
		else:
			return sp.ndimage.gaussian_filter(
				(self._components_best if self.give_best_result else self._components),
				self.clean_beam_sigma
			)
	def get_residual(self) -> np.ndarray:
		#return(self._residual)
		return self._obs - sp.signal.fftconvolve(self.get_components(), self._psf, mode='same')/self._flux_correction_factor
	def get_iters(self):
		return(self._i_best if self.give_best_result else self._i)
	
	
	def _init_algorithm(self, obs, psf) -> None:
		super(CleanModified, self)._init_algorithm(obs, psf)
		# initialise arrays
		self._obs = obs
		self._psf = psf
		self._components = np.zeros_like(obs)
		self._components_best = np.zeros_like(obs)
		
		self._i_best = 0
		self._residual = np.array(obs)
		self._residual_copy = np.array(self._residual)
		
		#self._iter_stat_names = ('fabs', 'rms', 'UNUSED', 'UNUSED', 'UNUSED','generalised_least_squares', 'UNUSED', 'UNUSED')
		#self._iter_stat_record = np.full((self.n_iter, 8), fill_value=np.nan)
		
		self._iter_stat_names = ('fabs', 'rms', 'generalised_least_squares')
		self._iter_stat_record = np.full((self.n_iter, len(self._iter_stat_names)), fill_value=np.nan)
		self._stats_best = np.full((len(self._iter_stat_names),),np.inf)
		
		self._selected_px = np.zeros_like(obs)
		self._selected_map = np.zeros_like(obs, dtype=bool)
		self._current_cleaned = np.zeros_like(obs)
		self._accumulator = np.zeros_like(obs)
		self._current_convolved = np.zeros_like(obs)
		self._current_cleaned = np.zeros_like(obs)
		self._tmp_r = np.zeros_like(obs)
		
		self._stats_delta = np.zeros((len(self._iter_stat_names),))
		#self._recency_weighted_rms_delta = 0
		#self._recency_weighted_fabs_delta = 0
		
		# this variable should only ever reference another array variable
		# and not have a value of it's own. We use it to change how we
		# pick the pixels to be adjusted without changing later code
		self._px_choice_img_ptr = Pointer(self._tmp_r)
		#self.testing_ptr = Pointer(np.zeros_like(obs))
		
		self._loop_gain_correction_factor = 1.0/np.nanmax(psf)
		self._flux_correction_factor = np.nansum(psf)
		
		self._choose_update_pixels = self.choose_residual_extrema
		#self.choose_update_pixels = self.choose_regularised_least_squares
		#self.choose_update_pixels = self.choose_generalised_least_squares
		#self.choose_update_pixels = self.choose_residual_extrema_reject_entropy
		
		# set variable values
		fabs = np.nanmax(np.fabs(self._residual))
		rms = np.sqrt(np.nansum(self._residual**2)/self._residual.size)
		self._fabs_threshold = fabs*self.fabs_frac_threshold
		self._rms_threshold = rms*self.rms_frac_threshold
		#self._fabs_min_delta = fabs*self.fabs_min_delta
		#self._rms_min_delta = rms*self.rms_min_delta

		# ensure that PSF is centreed and an odd number in shape
		#slices = tuple(slice(s-s%2) for s in psf.shape)
		#psf = psf[slices]
		#bp = np.unravel_index(np.nanargmax(psf), psf.shape)
		#ut.np.centre_on(psf, np.array(bp))
		self._pixel_threshold = 0


		if self.threshold < 0:
			#self.get_pixel_threshold = lambda : im_proc.otsu_thresholding.n_thresholds(self.residual_copy, 4)[-1]
			self._get_pixel_threshold = lambda : im_proc.otsu_thresholding.max_frac_diff_threshold(self._residual_copy)
		else:
			self._get_pixel_threshold = lambda : self.threshold*np.nanmax(self._px_choice_img_ptr.val)
			#self._get_pixel_threshold = lambda : self.threshold*np.nanmax(sp.ndimage.gaussian_filter(self._px_choice_img_ptr.val, sigma=2))

		return
	
	
	
	def _stopping_criteria(self):
	
		# Check thresholds
		_lgr.debug(f'{self._rms_threshold=:0.3g} rms = {self._iter_stat_record[self._i,1]:0.3g}')
		_lgr.debug(f'{self._fabs_threshold=:0.3g} fabs = {self._iter_stat_record[self._i,0]:0.3g}')
		if (self._iter_stat_record[self._i,1] < self._rms_threshold):
			self.progress_string = f"Ended at {self._i} iterations: Root mean squared statistic of residual has dropped below set threshold defined in `rms_frac_threshold` parameter."
			return(False)
			
		if (self._iter_stat_record[self._i,0] < self._fabs_threshold):
			self.progress_string = f"Ended at {self._i} iterations: Absolute brightest pixel value of residual has dropped below set threshold defined in `fabs_frac_threshold` parameter."
			return(False)
	
	
		# Check each statistic is better than the previously saved values
		new_best = True
		for name, this_stat, best_stat in zip(self._iter_stat_names, self._iter_stat_record[self._i], self._stats_best):
			if name != 'UNUSED':
				_lgr.debug(f'{new_best=} {this_stat=} {best_stat=}')
				new_best &= this_stat <= best_stat
		# If all current statistics are better, save the current state. 
		# Else, work out the maximum fractional increase . If it's 
		# larger than a threshold terminate iteration and return the best saved result
		if new_best:
			self._components_best[...] = self._components
			self._i_best = self._i
			self._stats_best = self._iter_stat_record[self._i]
		else:
			increase_frac = (self._iter_stat_record[self._i] - self._stats_best)/self._stats_best
			increase_frac_idxs_above_threshold = np.argwhere(increase_frac > self.max_stat_increase)
			_lgr.debug(f'{increase_frac=}')
			if increase_frac_idxs_above_threshold.size > 0:
				self.progress_string = f"Ended at {self._i} iterations: Statistics {self._iter_stat_names[increase_frac_idxs_above_threshold]} have increased from the best seen statistic beyond set threshold defined by `max_stat_increase` parameter."
				return False
		
		
		
		# Check the stability of the statistics. If they have all stopped evolving, terminate iteration
		n_lookback = 10
		if self._i >=n_lookback :
			for j, name in enumerate(self._iter_stat_names):
				if name != 'UNUSED':
					self._stats_delta[j] = np.std(self._iter_stat_record[self._i-n_lookback:self._i,j])/self._iter_stat_record[self._i,j]
					_lgr.debug(f'{name} fractional standard deviation {self._stats_delta[j]}')
			_lgr.debug(f'{self.min_frac_stat_delta=}')
			if all([_std < self.min_frac_stat_delta for j, _std in enumerate(self._stats_delta) if self._iter_stat_names[j] != 'UNUSED']):
				self.progress_string = f"Ended at {self._i} iterations: Standard deviation of statistics in last {n_lookback} steps are all below minimum fraction as defined in `min_frac_stat_delta` parameter."
				return False
				
		return True
	
	def _iter(self, obs, psf) -> bool:
		if not super(CleanModified, self)._iter(obs, psf): return(False)
		self._selected_px[...] *= 0 # reset selected pixels at start of iteration
		
		self._choose_update_pixels() # sets "self.choice_img_ptr" to an array that holds the values we should use for choosing pixels
		self._pixel_threshold = self._get_pixel_threshold()
		self._selected_map[...] = (self._px_choice_img_ptr.val > self._pixel_threshold)
		
		#min_region_size = 5
		#self._selected_map[...] = (sp.ndimage.grey_closing(self._px_choice_img_ptr.val, size=(min_region_size,min_region_size)) > self._pixel_threshold)
		#self._selected_map[...] = (sp.ndimage.gaussian_filter(self._px_choice_img_ptr.val, sigma=2) > self._pixel_threshold)
		
		
		# erode then dilate chosen pixels to remove single pixels, algorithm must have regions
		# of at least the size of the PSF to operate upon
		#for i in range(min_region_size):
		#	self._selected_map = sp.ndimage.binary_closing(self._selected_map, iterations=min_region_size//2)
		#self._selected_map = sp.ndimage.binary_dilation(self._selected_map, iterations=min_region_size//2)
		#self._selected_map = sp.ndimage.binary_erosion(self._selected_map, iterations=min_region_size)
		#self._selected_map = sp.ndimage.binary_dilation(self._selected_map, iterations=min_region_size)
		#self._selected_map = sp.ndimage.gaussian_filter(self._selected_map.astype(float), sigma=2) > 0.01
		
		"""
		label_map, n_labels = sp.ndimage.label(self._selected_map)
		# order labels by number of pixels they contain
						
		ordered_labels = list(range(1,n_labels+1)) # label 0 is background
		ordered_labels.sort(key = lambda x: np.count_nonzero(label_map == x), reverse=True)
		
		self._selected_map *= False
		for i in range(n_labels):
			if np.count_nonzero(label_map == ordered_labels[i]) >= min_region_size:
				self._selected_map |= (label_map == ordered_labels[i])
		"""
		
		
		
		self._selected_px[self._selected_map] = self._residual[self._selected_map]*self.loop_gain
		
		#self._selected_px[...] = sp.ndimage.median_filter(self._selected_px, size=5)
		#self._selected_px[...] = sp.ndimage.gaussian_filter(self._selected_px, sigma=1)
		
		# TESTING DIFFERENT STRATEGIES
		# This one may have better convergence statistics, I should check it
		#rma = np.nanmax(np.abs(self._residual))
		#r2 = (np.abs(self._residual)/rma)**2
		#self._selected_px[self._selected_map] = (np.sign(self._residual)*(r2*rma))[self._selected_map]*self.loop_gain
		self._accumulator[self._selected_map] += 1
		
		# convolve selected pixels with PSF and adjust so that flux is conserved
		self._current_convolved[...] = sp.signal.fftconvolve(self._selected_px, psf, mode='same')/self._flux_correction_factor
		
		# update residual
		self._residual[...] -= self._current_convolved
		self._residual_copy[...] = self._residual
		
		# add new components to component map
		self._components[...] += self._selected_px
		
		# update the current cleaned map
		self._current_cleaned[...] += self._current_convolved
		
		# update statistics that tell us how it's going
		self._iter_stat_record[self._i] = (	
			np.nanmax(np.fabs(self._residual)), # fabs
			np.sqrt(np.nansum(self._residual**2)/self._residual.size), # rms
			np.nansum( # generalised_least_squares
				0.5*mfunc.generalised_least_squares(
					self._components,
					obs, self.noise_std, psf
				)
			),
		)
		
		return self._stopping_criteria()


	def choose_residual_extrema(self):
		if (self._i < self.n_positive_iter):
			self._px_choice_img_ptr.val = self._residual
		else:
			self._tmp_r[...] = np.fabs(self._residual)
			self._px_choice_img_ptr.val = self._tmp_r
		
		return



"""
Tool for fitting a PSF to a particular model
"""

import sys
from pathlib import Path
import datetime as dt
from functools import partial


import numpy as np
import scipy as sp
from astropy.io import fits

import aopp_deconv_tool.numpy_helper as nph

import aopp_deconv_tool.astropy_helper as aph
import aopp_deconv_tool.astropy_helper.fits.specifier
import aopp_deconv_tool.astropy_helper.fits.header
from aopp_deconv_tool.fpath import FPath
import aopp_deconv_tool.plot_helper as plot_helper
import aopp_deconv_tool.arguments


from aopp_deconv_tool.optimise_compat import PriorParam, PriorParamSet
from aopp_deconv_tool.optimise_compat.ultranest_compat import UltranestResultSet, fitting_function_factory


from aopp_deconv_tool.psf_model.dependency_injector import (
	RadialPSFModelDependencyInjector, 
	GaussianPSFModelDependencyInjector,
	TurbulencePSFModelDependencyInjector,
	MUSEAdaptiveOpticsPSFModelDependencyInjector,	
)

import aopp_deconv_tool.psf_data_ops as psf_data_ops

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')



psf_models = {
	'radial' : RadialPSFModelDependencyInjector,
	'gaussian' : GaussianPSFModelDependencyInjector,
	'tubulence' : TurbulencePSFModelDependencyInjector,
	'muse_ao' : MUSEAdaptiveOpticsPSFModelDependencyInjector,
}

FITTING_METHODS=('ultranest', 'scipy.minimize')

PSF_MODEL_DI_CLASS = None
PSF_MODEL_DI = None

def set_psf_model_dependency_injector(name, fits_spec):
	global PSF_MODEL_DI_CLASS
	global PSF_MODEL_DI
	PSF_MODEL_DI_CLASS = psf_models[name]
	with fits.open(Path(fits_spec.path)) as data_hdul:
		with nph.axes.to_end(data_hdul[fits_spec.ext].data[fits_spec.slices], fits_spec.axes['CELESTIAL']) as data:
			PSF_MODEL_DI = PSF_MODEL_DI_CLASS(data[tuple(0 for i in range(data.ndim-2))])
	return

def get_psf_model_dependency_injector():
	return PSF_MODEL_DI

def get_new_psf_model_dependency_injector(data):
	di = PSF_MODEL_DI_CLASS(data)
	di._params = PriorParamSet(*PSF_MODEL_DI._params.prior_params)
	return di



def set_psf_model_dependency_injector_params(variables=[], **kwargs):
	global PSF_MODEL_DI
	new_params = []
	_lgr.debug(f'{variables=}')
	_lgr.debug(f'{kwargs=}')
	for p in PSF_MODEL_DI._params.prior_params:
		new_params.append(p)
		new_params[-1].const_value = kwargs.get(p.name, p.const_value)
		new_params[-1].domain = kwargs.get(p.name+'_domain', p.domain)
		new_params[-1].is_const = p.name not in variables
	
	PSF_MODEL_DI._params = PriorParamSet(*new_params)
		
		

	
def plot_result(result, psf_data, suptitle=None, show=True):
	if not show: return

	residual = psf_data-result
	
	f, a = plot_helper.figure_n_subplots(7)
	
	a[0].set_title(f'data sum={np.nansum(psf_data)}')
	a[0].imshow(np.log(psf_data))
	a[0].plot([psf_data.shape[0]/2],[psf_data.shape[1]/2], 'r.')
	
	a[1].set_title(f'result sum={np.nansum(result)}')
	a[1].imshow(np.log(result))
	a[1].plot([result.shape[0]/2],[result.shape[1]/2], 'r.')
	
	a[2].set_title(f'residual sum={np.nansum(residual)}')
	a[2].imshow(np.log(residual))
	
	a[3].set_title(f'residual_squared sqrt(sum)={np.sqrt(np.nansum(residual**2))}')
	a[3].imshow(np.log((residual)**2))
	
	a[4].set_title('data and result slice (horizontal)')
	a[4].plot(np.log(psf_data[psf_data.shape[0]//2,:]).flatten())
	a[4].plot(np.log(result[result.shape[0]//2,:]).flatten())
	a[4].axvline(result.shape[0]//2, color='red', ls='--')

	
	a[5].set_title('data and result slice (vertical)')
	a[5].plot(np.log(psf_data[:,psf_data.shape[1]//2]).flatten())
	a[5].plot(np.log(result[:,result.shape[1]//2]).flatten())
	a[5].axvline(result.shape[1]//2, color='red', ls='--')
	
	offsets_from_centre = nph.array.offsets_from_point(psf_data.shape)
	offsets_from_centre = (offsets_from_centre.T - np.array([0.0,0.5])).T
	r_idx1 = np.sqrt(np.sum(offsets_from_centre**2, axis=0))
	r = np.linspace(0,np.max(r_idx1),30)
	psf_radial_data = np.array([np.nansum(psf_data[(r_min <= r_idx1) & (r_idx1 < r_max) ]) for r_min, r_max in zip(r[:-1], r[1:])])
	
	offsets_from_centre = nph.array.offsets_from_point(psf_data.shape)
	offsets_from_centre = (offsets_from_centre.T - np.array([0.5,0.5])).T
	r_idx2 = np.sqrt(np.sum(offsets_from_centre**2, axis=0))
	r = np.linspace(0,np.max(r_idx2),30)
	result_radial_data = np.array([np.nansum(result[(r_min <= r_idx2) & (r_idx2 < r_max) ]) for r_min, r_max in zip(r[:-1], r[1:])])
	
	a[6].set_title('radial data and result')
	a[6].plot(r[:-1], psf_radial_data)
	a[6].plot(r[:-1], result_radial_data)
	
	f.suptitle(suptitle)
	
	plot_helper.output(show)


def run(
		fits_spec,
		output_path,
		fit_result_dir : str | None = None,
		method : str = 'ultranest',
	):
	
	original_data_type=None
	
	if fit_result_dir is None:
		fit_result_dir = Path(output_path).parent
	
	axes = fits_spec.axes['CELESTIAL']
	
	with fits.open(Path(fits_spec.path)) as data_hdul:
		
		
		_lgr.debug(f'{fits_spec.path=} {fits_spec.ext=} {fits_spec.slices=} {fits_spec.axes=}')
		#raise RuntimeError(f'DEBUGGING')
	
		data_hdu = data_hdul[fits_spec.ext]
		data = data_hdu.data
		hdr = data_hdu.header
		original_data_type = data_hdu.data.dtype
		axes = fits_spec.axes['CELESTIAL']
		spectral_axes = aph.fits.header.get_spectral_axes(hdr)
		if len(spectral_axes) != 1:
			raise RuntimeError('Data must have one spectral axis')
		
		
		spectral_axes_slices = tuple(slice(None) if x in spectral_axes else 0 for x in range(data.ndim))
		_lgr.debug(f'{spectral_axes=}')
		
		spectral_coords = aph.fits.header.get_world_coords_of_axis(hdr, spectral_axes)
		_lgr.debug(f'{spectral_coords.shape=}')
		
		_lgr.debug(f'{PSF_MODEL_DI._params=}')
		
		timestamp = dt.datetime.now(dt.timezone.utc).isoformat()
		result_set_directory = fit_result_dir / f'{timestamp}_psf_fit_result'
			
		match method:
			case 'ultranest':
				result_set = UltranestResultSet(Path(result_set_directory))
				result_set.metadata['wavelength_idxs'] = tuple((x,i) for i,x in enumerate(spectral_coords))
				result_set.save_metadata()
		
		result_callables = []
		
		result_data = np.full_like(data, np.nan)
		fitted_params = [None]*spectral_coords.shape[0]
		
		for i, idx in enumerate(nph.slice.iter_indices(data, fits_spec.slices, axes)):
			_lgr.debug(f'{spectral_axes_slices=}')
			j = idx[spectral_axes[0]]
			for _ in range(len(idx)-1):
				j = j[0]
			_lgr.debug(f'{i=} {j=}')
			
			psf_data = data[idx] #np.nan_to_num(data[idx])
			psf_err = 1E-2*psf_data + 1E-3*np.nanmax(psf_data)
			di = get_new_psf_model_dependency_injector(psf_data)
			
			if 'wavelength' in di._params.all_params:
				wavelength = spectral_coords[j]
				variables = [x.name for x in di._params.variable_params if x.name != wavelength]
				set_psf_model_dependency_injector_params(variables, {'wavelength' : wavelength})
	
	
			psf_model_name = di.get_psf_model_name()
			params = di.get_parameters()
			psf_model_callable = di.get_psf_model_flattened_callable()
			psf_result_postprocess = di.get_psf_result_postprocessor()
			result_callables.append(di.get_fitted_parameters_callable())
			
			
	
			

			# Get correct fitting function and objective function for fitting method
			fitting_function = None
			objective_function_factory = None
			
			match method:
				case 'ultranest':
					# NOTE: Alter the parameters here to make the nested sampling more exact/take longer vs less exact/run faster
					num_live_points = 20
					result_set.metadata['constant_parameters'] = [p.to_dict() for p in params.constant_params]
					result_set.save_metadata()
					fitting_function = fitting_function_factory(
						reactive_nested_sampler_kwargs = {
							'log_dir' : result_set.directory,
							'run_num' : j
						},
						sampler_run_kwargs = {
							'max_iters' : 500, #500, # 2000,
							'max_ncalls' : 5000, #5000
							'frac_remain' : 1E-2,
							'Lepsilon' : 1E-1,
							'min_num_live_points' : num_live_points, #20, #80
							'cluster_num_live_points' : num_live_points//5, #1, #40
							'min_ess' : num_live_points, #1, #40
							'widen_before_initial_plateau_num_warn' : (3*num_live_points)//2, #*min_live_points,
							'widen_before_initial_plateau_num_max' : 2*num_live_points #*min_live_points
						}
					)
					objective_function_factory = partial(psf_data_ops.objective_function_factory, mode='maximise')
				case 'scipy.minimize':
					fitting_function = psf_data_ops.scipy_fitting_function_factory(sp.optimize.minimize)
					objective_function_factory = partial(psf_data_ops.objective_function_factory, mode='minimise')
				case _:
					raise RuntimeError(f'Unknown fitting method "{method}". Should be one of {FITTING_METHODS}')
			
			
			# Perform fit
			fitted_psf, fitted_vars, consts = psf_data_ops.fit_to_data(
				params, 
				psf_model_callable, 
				psf_data, 
				psf_err,
				fitting_function,
				objective_function_factory,
				plot_mode=None
			)
			_lgr.info(f'{fitted_vars=}')
			_lgr.info(f'{consts=}')
			
			fitted_params[j] = {**fitted_vars, **consts}
			
			# Do any postprocessing if we need to
			if psf_result_postprocess is not None:
				result_data[idx] = psf_result_postprocess(params, psf_model_callable, fitted_vars, consts)
				plot_result(result_data[idx], psf_data, suptitle=f'{j=}\n{fitted_vars=}', show=False)
			
			
		
		match method:
			case 'ultranest':
				result_set.plot_params_vs_run_index(show=False, save=True)
				result_set.plot_results(
					result_callables,
					data,
					show=False,
					save=True
				)
		
		
	param_dict = {
		'method' : method,
		'model' : PSF_MODEL_DI._psf_model.__class__.__name__,
		'fit_result_dir' : None if (method not in ('ultranest',)) else fit_result_dir
	}
	
	hdr.update(aph.fits.header.DictReader(
		param_dict,
		prefix='fit_psf_model',
		pkey_count_start=aph.fits.header.DictReader.find_max_pkey_n(hdr)
	))
	


	_lgr.info('save the products to a FITS file')
	hdus = []

	hdus.append(fits.PrimaryHDU(
		header = hdr,
		data = result_data.astype(original_data_type)
	))
	
	
	
	hdus.append(fits.BinTableHDU.from_columns(
		columns = [
			fits.Column(name=k, format='D', array=[x[k] if x is not None else np.nan for x in fitted_params]) for k in fitted_params[[i for i,x in enumerate(fitted_params) if x is not None][0]].keys()
		],
		name = 'FITTED_MODEL_PARAMS',
		header = None,
	))
	
	hdul_output = fits.HDUList(hdus)
	hdul_output.writeto(output_path, overwrite=True)


def parse_args(argv):
	import os
	import aopp_deconv_tool.text
	import argparse
	
	DEFAULT_OUTPUT_TAG = '_modelled'
	DESIRED_FITS_AXES = ['CELESTIAL']
	OUTPUT_COLUMNS=80
	try:
		OUTPUT_COLUMNS = os.get_terminal_size().columns - 30
	except Exception:
		pass
	
	FITS_SPECIFIER_HELP = aopp_deconv_tool.text.wrap(
		aph.fits.specifier.get_help(DESIRED_FITS_AXES).replace('\t', '    '),
		OUTPUT_COLUMNS
	)
	
	class ArgFormatter (argparse.RawTextHelpFormatter, argparse.ArgumentDefaultsHelpFormatter, argparse.MetavarTypeHelpFormatter):
		def __init__(self, *args, **kwargs):
			super().__init__(*args, **kwargs)
	
	parser = argparse.ArgumentParser(
		description=__doc__, 
		formatter_class=ArgFormatter,
		epilog=FITS_SPECIFIER_HELP
	)
	
	parser.add_argument(
		'fits_spec',
		help = '\n'.join((
			f'FITS Specifier of the data to operate upon . See the end of the help message for more information',
			f'required axes: {", ".join(DESIRED_FITS_AXES)}',
		)),
		type=str,
		metavar='FITS Specifier',
	)
	
	#parser.add_argument('-o', '--output_path', type=str, help=f'Output fits file path. If None, is same as the `fits_spec` path with "{DEFAULT_OUTPUT_TAG}" appended to the filename')
	parser.add_argument(
		'-o', 
		'--output_path', 
		type=FPath,
		metavar='str',
		default='{parent}/{stem}{tag}_{model}{suffix}',
		help = '\n    '.join((
			f'Output fits file path, supports keyword substitution using parts of `fits_spec` path where:',
			'{parent}: containing folder',
			'{stem}  : filename (not including final extension)',
			f'{{tag}}   : script specific tag, "{DEFAULT_OUTPUT_TAG}" in this case',
			'{model} : model fitted to psf',
			'{suffix}: final extension (everything after the last ".", including the ".")',
			'\b'
		))
	)
	
	parser.add_argument('--fit_result_dir', type=str, default=None, help='Directory to store results of PSF fit in. Will create a sub-directory below the given path. If None, will create a sibling folder to the output file (i.e., output file parent directory is used).')

	parser.add_argument('--model', type=str, default='radial', choices=tuple(psf_models.keys()), help='Model to fit to PSF data.')

	parser.add_argument('--method', type=str, default='scipy.minimize', choices=FITTING_METHODS, help='What method should we use to perform the fitting')

	parser.add_argument('--model_help', action='store_true', default=False, help='Show the help message for the selected model')

	args, psf_model_args = parser.parse_known_args(argv)
	
	args.fits_spec = aph.fits.specifier.parse(args.fits_spec, DESIRED_FITS_AXES)
	
	#if args.output_path is None:
	#	args.output_path =  (Path(args.fits_spec.path).parent / (str(Path(args.fits_spec.path).stem)+DEFAULT_OUTPUT_TAG+f'_{args.model}'+str(Path(args.fits_spec.path).suffix)))
	other_file_path = Path(args.fits_spec.path)
	args.output_path = args.output_path.with_fields(
		tag=DEFAULT_OUTPUT_TAG, 
		parent=other_file_path.parent, 
		stem=other_file_path.stem, 
		suffix=other_file_path.suffix,
		model=args.model,
	)
	
	set_psf_model_dependency_injector(args.model, args.fits_spec)
	
	di = get_psf_model_dependency_injector()
	params = di._params.all_params
	
	di_param_parser = argparse.ArgumentParser(
		prog=f'fit_psf_model.py --model {args.model}',
		description=get_psf_model_dependency_injector().__doc__, 
		formatter_class=ArgFormatter,
		add_help=False,
	)
	
	for item in params:
		di_param_parser.add_argument("--"+item.name, type=float, default=di._params[item.name].const_value, help=item.description)
		di_param_parser.add_argument("--"+item.name+'_domain', nargs=2, type=float, default=di._params[item.name].domain, help=f'domain parameter "{item.name}" for "{args.model}" psf model')
	
	default_variable_params = [x.name for x in di._params.variable_params]
	di_param_parser.add_argument('--variables', nargs='*', type=str, default=default_variable_params, choices=[x.name for x in params], help=f'Which parameters to vary when fitting, others will be held constant.')
	
	if args.model_help:
		di_param_parser.print_help()
		sys.exit()
	
	di_params = vars(di_param_parser.parse_args(psf_model_args))
	_lgr.debug(f'{di_params["variables"]=}')
	
	set_psf_model_dependency_injector_params(**di_params)
	
	
	
	
	
	
	
	
	return args

def go(
		fits_spec,
		output_path=None,
		fit_result_dir=None,
		model=None,
		method=None,
		model_help=None
	):
	"""
	Thin wrapper around `run()` to accept string inputs.
	As long as the names of the arguments to this function 
	are the same as the names expected from the command line
	we can do this programatically
	"""
	# Add stuff to kwargs here if needed
	
	# This must be first so we only grab the arguments to the function
	fargs = dict(locals().items())
	arglist = aopp_deconv_tool.arguments.construct_arglist_from_locals(fargs, n_positional_args=1)
	exec_with_args(arglist)

def exec_with_args(argv):
	args = parse_args(argv)
	
	run(
		args.fits_spec, 
		args.output_path,
		args.fit_result_dir,
		args.method,
	)


if __name__=='__main__':
	exec_with_args(sys.argv[1:])




from typing import Optional, Literal, Any
import dataclasses as dc

import numpy as np
import scipy as sp

from .base import Base
import aopp_deconv_tool.numpy_helper as nph
import aopp_deconv_tool.numpy_helper.array.pad
import aopp_deconv_tool.numpy_helper.slice
import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')
	
@dc.dataclass(slots=True, repr=False, eq=False,)
class LucyRichardson(Base):
	"""
	Implementation of the Lucy-Richardson algorithm
	"""
	
	# Public attributes
	
	nudge_factor 	: float = dc.field(default=1E-2, 	metadata={
		'description':'Fraction of maximum brightness to add to numerator and denominator to try and avoid numerical instability. This value should be in the range [0,1), and will usually be small. Larger values require more steps to give a solution, but suffer less numerical instability.',
		'domain' : (0,1),
	})
	strength 		: float = dc.field(default=1E-1,	metadata={
		'description':'Multiplier to the correction factors, if numerical insability is encountered decrease this. A more crude method of avoiding instability than `nudge_factor`, should be in the range (0,1].',
		'domain':(0,1),
	})
	cf_negative_fix : bool 	= dc.field(default=True, 	metadata={
		'description':'Should we change negative correction factors to close-to-zero correction factors? Usually we should as we don\'t want any negative correction factors to flip-flop the end result.',
		'domain' : (False,True),
	})
	cf_limit 		: float = dc.field(default=np.inf,	metadata={
		'description':'End iteration if the correction factors are larger than this limit. Large correction factors are a symptom of numerical instability.',
		'domain' : (0, np.inf),
	})
	cf_uclip 		: float = dc.field(default=np.inf,	metadata={
		'description':'Clip the correction factors to be no larger than this value. A crude method to control numerical instability.',
		'domain' : (-np.inf,np.inf),
	})
	cf_lclip 		: float = dc.field(default=-np.inf,	metadata={
		'description':'Clip the correction factors to be no smaller than this value. A crude method to control numerical instability.',
		'domain' : (-np.inf, np.inf),
	})
	offset_obs 		: bool 	= dc.field(default=False,	metadata={
		'description':'Should we offset the observation so there are no negative pixels? Enables the algorithm to find -ve values as the offset is reversed at the end.',
		'domain' : (False, True),
	})
	threshold : \
		Optional[float]     = dc.field(default=None,	metadata={
		'description':'Below this value LR will not be applied to pixels. This is useful as at low brightness LR has a tendency to fit itself to noise. If -ve will use |threshold|*brightest_pixel as threshold each step. If zero will use mean and standard deviation to work out a threshold, if None will not be used.',
		'domain' : (-np.inf, np.inf),
	})
	pad_observation : bool  = dc.field(default=True,	metadata={
		'description':'Should we pad the input data with extra space to avoid edge effects? Padding will take the form of convolving the observation with the psf, but only keeping the edges of the convolved data. This will hopefully cause a smooth-ish drop-off at the edges instead of a hard cutoff, thus reducing insability.',
		'domain' : (False, True),
	})
	
	# Private attributes
	#_example : np.ndarray = dc.field(init=False, repr=False, hash=False, compare=False)
	_nudge : float = dc.field(init=False, repr=False, hash=False, compare=False)
	_obs_shape : tuple[int,...] = dc.field(init=False, repr=False, hash=False, compare=False)
	_dirty_img : float = dc.field(init=False, repr=False, hash=False, compare=False)
	_offset : float = dc.field(init=False, repr=False, hash=False, compare=False)
	_psf_reversed : np.ndarray = dc.field(init=False, repr=False, hash=False, compare=False)
	_cf : np.ndarray = dc.field(init=False, repr=False, hash=False, compare=False)
	_blurred_est : np.ndarray = dc.field(init=False, repr=False, hash=False, compare=False)
	_obs_per_est : np.ndarray = dc.field(init=False, repr=False, hash=False, compare=False)
	_out_of_bounds_mask : np.ndarray = dc.field(init=False, repr=False, hash=False, compare=False)
	
	

	def _init_algorithm(self, obs, psf):
		super(LucyRichardson, self)._init_algorithm(obs, psf)
		self._nudge = self.nudge_factor*np.max(obs) if self.nudge_factor > 0 else -self.nudge_factor

		# Pad observation to avoid edge effects
		self._obs_shape = obs.shape
		self._dirty_img = nph.array.pad.with_convolution(obs, psf) if self.pad_observation else np.array(obs)
		
		# offset dirty_img if we want to allow -ve values in our results
		# increase the value from 0 to allow more -ve values
		self._offset = 0-np.nanmin(self._dirty_img) if self.offset_obs else 0
		self._dirty_img += self._offset

		# initialise common arrays
		self._components = np.ones_like(self._dirty_img)*np.mean(self._dirty_img)
		self._residual = np.array(self._dirty_img)

		# initialise special arrays
		self._psf_reversed = np.array(psf[::-1,::-1])
		self._cf = np.ones_like(self._dirty_img) # correction factors
		self._blurred_est = np.zeros_like(self._components)
		self._obs_per_est = np.zeros_like(self._components)
		self._out_of_bounds_mask = np.ones_like(self._components, dtype=bool)
	
		# set array values that can't be done during initalisation	
		self._out_of_bounds_mask[nph.slice.around_centre(self._out_of_bounds_mask.shape, obs.shape)] = 0

	
		return

	def get_components(self):
		if self.pad_observation:
			return(self._components[nph.slice.around_centre(self._dirty_img.shape, self._obs_shape)] - self._offset)
		else:
			return(self._components - self._offset) 
	
	def get_residual(self):
		if self.pad_observation:
			return(self._residual[nph.slice.around_centre(self._dirty_img.shape, self._obs_shape)])
		else:
			return(self._residual) 
	
	def get_iters(self):
		return self._i
	
	def _iter(self, obs, psf):
		"""
		Perform an iteration of the algorithm, return True if iteration was successful, False otherwise
		"""
		if not super(LucyRichardson, self)._iter(obs, psf): return(False)
		self._blurred_est[...] = sp.signal.fftconvolve(self._components, psf, mode='same')
		#self.blurred_est[self.blurred_est==0] = np.min(np.fabs(self.blurred_est))

		self._obs_per_est[...] = (self._dirty_img + self._nudge)/(self._blurred_est + self._nudge)

		self._cf[...] = sp.signal.fftconvolve(self._obs_per_est, self._psf_reversed, mode='same')
		self._cf[...] = self.strength*(self._cf - 1) + 1 # correction factors should always be centreed around one

		# if we have a threshold, apply it
		if self.threshold is not None:
			if self.threshold < 0:
				threshold_value = np.nanmax(self._residual)*abs(self.threshold)
			elif self.threshold == 0:
				threshold_value = np.nanmean(self._residual)+1*np.nanstd(self._residual)
			else:
				threshold_value = self.threshold
			self._cf[self._residual < threshold_value] = 1

		# once these get large the result becomes unstable, clip the upper bound if desired
		if (self.cf_uclip != np.inf):
			self._cf[self._cf>self.cf_uclip] = self.cf_uclip
		
		# anything close to zero can just be zero, clip the lower bound if desired
		if (self.cf_lclip != -np.inf):
			self._cf[self.cf<self.cf_lclip] = 0
			
		# we probably shouldn't even have -ve correction factors, turn them into a close-to-zero factor instead
		
		if self.cf_negative_fix:
			cf_negative = self._cf < 0
			if np.any(cf_negative):
				cf_positive = self._cf > 0
				if not np.any(cf_positive):
					raise ValueError('All correction factors in Lucy-Richardson deconvolution have become negative, exiting...')
				self._cf[cf_negative] = np.min(self._cf[cf_positive])*np.exp(self._cf[cf_negative])
		
		self._components *= self._cf
		self._components[self._out_of_bounds_mask] = self._offset
 
		# could make this faster by doing an inexact version
		self._residual[...] = self._dirty_img - sp.signal.fftconvolve(self._components, psf, mode='same')
		#self.resigual = self.dirty_img - self.blurred_est
	
		# If any of our exit conditions trip, exit the loop
		if np.nanmax(np.fabs(self._cf)) > self.cf_limit:
			print('WARNING: Correction factors getting large, stopping iteration')
			return(False)
		if np.all(np.isnan(self._cf)):
			print('ERROR: Correction factors have all become NAN, stopping iteration')
			return(False)
		return(True)



"""
Apply slices to a FITS file, reducing the number of entries across a set of axes.
"""


import sys, os
from pathlib import Path


from astropy.io import fits

import aopp_deconv_tool.astropy_helper as aph
import aopp_deconv_tool.astropy_helper.fits.header
import aopp_deconv_tool.astropy_helper.fits.specifier
from aopp_deconv_tool.astropy_helper.fits.specifier import FitsSpecifier

from aopp_deconv_tool.fpath import FPath
from aopp_deconv_tool.numpy_helper.axes import AxesOrdering
import aopp_deconv_tool.arguments

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')



def run(
		fits_spec : FitsSpecifier,
		output_path : Path|str,
		#squeeze : bool = False
	):
	
	with fits.open(Path(fits_spec.path)) as hdul:
		hdu = hdul[fits_spec.ext]
		original_data_type = hdu.data.dtype
		original_data_shape = hdu.data.shape
		data = hdu.data[fits_spec.slices]
		assert data.size > 0, f"Slice selects no data elements: slice={fits_spec.slices}; data_shape={original_data_shape}"
		_lgr.debug(f'{data.shape=}')
		hdr = hdu.header
		hdr.update(aph.fits.header.DictReader(
			{
				'original_file' : fits_spec.path,
			},
			prefix='slice',
			pkey_count_start=aph.fits.header.DictReader.find_max_pkey_n(hdr)
		))
	
	for axis in fits_spec.axes['ALL']:
		ax = AxesOrdering(axis, hdr['NAXIS'], 'numpy')
		if ax.numpy >= len(fits_spec.slices):
			continue
		
		aph.fits.header.set_axes_transform(
			hdr,
			axis = ax.fits,
			reference_pixel = int(hdr[f'CRPIX{ax.fits}']) - fits_spec.slices[ax.numpy].start
		)
		
		
	
	
	hdu_sliced = fits.PrimaryHDU(
		header = hdr,
		data = data.astype(original_data_type)
	)
	hdul_output = fits.HDUList([
		hdu_sliced,
	])
	hdul_output.writeto(output_path, overwrite=True)


def parse_args(argv):
	import os
	import aopp_deconv_tool.text
	import argparse
	
	DEFAULT_OUTPUT_TAG = '_sliced'
	DESIRED_FITS_AXES = ['ALL']
	FITS_SPECIFIER_HELP = aopp_deconv_tool.text.wrap(
		aph.fits.specifier.get_help(DESIRED_FITS_AXES).replace('\t', '    '),
		os.get_terminal_size().columns - 30
	)
	
	class ArgFormatter (argparse.RawTextHelpFormatter, argparse.ArgumentDefaultsHelpFormatter, argparse.MetavarTypeHelpFormatter):
		def __init__(self, *args, **kwargs):
			super().__init__(*args, **kwargs)
	
	parser = argparse.ArgumentParser(
		description=__doc__, 
		formatter_class=ArgFormatter,
		epilog=FITS_SPECIFIER_HELP,
		exit_on_error=False
	)
	
	parser.add_argument(
		'fits_spec',
		help = 'FITS file specifier, will slice the file as requested here',
		type=str,
		metavar='FITS Specifier',
	)
	
	#parser.add_argument(
	#	'--squeeze',
	#	help="If present, will remove any axes that end up with only 1 entry",
	#	action='store_true',
	#)
	
	parser.add_argument(
		'-o', 
		'--output_path', 
		type=FPath,
		metavar='str',
		default='{parent}/{stem}{tag}{suffix}',
		help = '\n    '.join((
			f'Output fits file path, supports keyword substitution using parts of `obs_fits_spec` path where:',
			'{parent}: containing folder',
			'{stem}  : filename (not including final extension)',
			f'{{tag}}   : script specific tag, "{DEFAULT_OUTPUT_TAG}" in this case',
			'{suffix}: final extension (everything after the last ".")',
			'\b'
		))
	)

	
	args = parser.parse_args(argv)
	
	args.fits_spec = aph.fits.specifier.parse(args.fits_spec, DESIRED_FITS_AXES)
	
	other_file_path = Path(args.fits_spec.path)
	args.output_path = args.output_path.with_fields(
		tag=DEFAULT_OUTPUT_TAG, 
		parent=other_file_path.parent, 
		stem=other_file_path.stem, 
		suffix=other_file_path.suffix
	)
	
	return args

def go(
		fits_spec,
		output_path=None
	):
	"""
	Thin wrapper around `run()` to accept string inputs.
	As long as the names of the arguments to this function 
	are the same as the names expected from the command line
	we can do this programatically
	"""
	# Add stuff to kwargs here if needed
	
	# This must be first so we only grab the arguments to the function
	fargs = dict(locals().items())
	arglist = aopp_deconv_tool.arguments.construct_arglist_from_locals(fargs, n_positional_args=1)
	exec_with_args(arglist)

def exec_with_args(argv):
	args = parse_args(argv)
	_lgr.info('#### ARGUMENTS ####')
	for k,v in vars(args).items():
		_lgr.info(f'\t{k} : {v}')
	_lgr.info('#### END ARGUMENTS ####')
	
	run(
		args.fits_spec, 
		args.output_path, 
		#squeeze = args.squeeze
	)
	
if __name__ == '__main__':
	exec_with_args(sys.argv[1:])

	


"""
Module containing small mathematical functions that are useful in lots of contexts



## Helpers for deconvolution routines ##

Working Definitions:
	components
		The pixel values of the underlying "real" image, i.e., our 'hypothetical' image
	model
		A "guess" as to what the underlying "real" image should be. Should be either
		a physically informed model (e.g., a synthetic image of Neptune) or the pixel
		values expected from an empty field (e.g., RMS noise values).
	estimate
		Our current estimate of the dirty_img (observation), is a convolution 
		between the PSF and the components
	dirty_img
		The observation we are trying to deconvolve
	psf
		The (or an estimate of) the point spread function that transforms the
		underlying "real" image into the dirty_img (observation)
	error
		The estimate of the error on the value of each pixel in the dirty_img 
		(observation). Should either be an RMS value, or a pixel map of values.
		Make the error for a pixel large if you want the solution to relax to the
		model values.
		
Useful Concepts:
	entropy
		A measure of how much extra information you have to add to a model.
		More information = more negative value.

"""

import numpy as np
import scipy as sp
import scipy.signal
from typing import Callable

Number = float | int


def logistic_function(
		x : Number | np.ndarray, 
		left_limit : Number = 0, 
		right_limit : Number = 1, 
		transition_scale : Number = 1, 
		centre : Number = 0
	) -> Number | np.ndarray:
	return (right_limit-left_limit)/(1+np.exp(-(np.e/transition_scale)*(x-centre))) + left_limit



# Different versions of generalised least squares

def generalised_least_squares(
		components 	: np.ndarray, 
		dirty_img 	: np.ndarray, 
		cov_mat 	: np.ndarray | float | int, 	# variances for the errors on data
		response 	: np.ndarray
		) -> np.ndarray:
		mismatch = sp.signal.fftconvolve(components, response, mode='same') - dirty_img
		return(mismatch*mismatch*(1.0/cov_mat))

def generalised_least_squares_preconv(
		estimate 	: np.ndarray, 
		dirty_img 	: np.ndarray, 
		cov_mat 	: np.ndarray | float | int, 	# variances for the errors on data
		) -> np.ndarray:
		return(((estimate - dirty_img)**2)*(1.0/cov_mat))
	
def generalised_least_squares_mat(
		components 	: np.matrix, 
		dirty_img 	: np.matrix, 
		cov_mat 	: np.matrix, 	# covariance matrix for the errors on data
		response 	: np.matrix
		) -> np.matrix:
		mismatch = response@components - dirty_img
		inv_cov_mat = np.linalg.inv(cov_mat)
		mm_matrix_row = mm_matrix if mm_matrix.shape[0] == 1 else mm_matrix.T
		return(np.einsum('ij,jk,kl->il', mm_matrix_row, inv_cov_mat, mm_matrix_row.T))


# Different versions of entropy

def entropy_pos_neg(
		components : np.ndarray,
		error : np.ndarray | float | int
		) -> np.ndarray:
	psi = np.sqrt(components**2 + 4*error**2)
	return(psi - 2*error - components*np.log((psi + components)/(2*error)))

def entropy_pos(
		components : np.ndarray,
		model : np.ndarray | float | int
		) -> np.ndarray:
	return(components - model - components*np.log(components/model))

def entropy_rel(
		components : np.ndarray,
		model : np.ndarray | float | int
		) -> np.ndarray:
	return(-components*np.log(components/model))

def entropy_adj(
		components : np.ndarray,
		model : np.ndarray | float | int,
		error : np.ndarray | float | int = 1 # this could be absorbed into alpha if we need to only have 2 arguments
		):
	return(entropy_pos_neg(components-model,error))



def regularised_least_squares(
		components 			: np.ndarray,
		dirty_img 			: np.ndarray,
		error 				: np.ndarray | float | int,
		alpha 				: np.ndarray | float | int, # no reason we can't have different values for different pixels
		response			: np.ndarray, # usually an instrumental PSF
		model 				: np.ndarray | float | int | None										= None,
		regularising_func 	: Callable[[np.ndarray, np.ndarray | float | int], np.ndarray] 	= entropy_adj,
		) -> np.ndarray:
	if model is None:
		model = error
	return(0.5*generalised_least_squares(components, dirty_img, error**2, response) - alpha*regularising_func(components, model))






import numpy as np
import scipy as sp
import scipy.ndimage

def quick_remove_nan_and_inf(data, size=5):
	mask = np.isnan(data) | np.isinf(data)
	if np.any(mask):
		#whole_median = np.nanmedian(data[~mask])
		data[mask] = 0
		#data[mask] = sp.ndimage.median_filter(data, size=size)[mask]
		data[mask] = sp.ndimage.gaussian_filter(data, sigma=size)[mask]
	return data


from aopp_deconv_tool.psf_model.dependency_injector.params_and_psf_model_di import *
from aopp_deconv_tool.psf_model.gaussian_psf_model import GaussianPSFModel


class GaussianPSFModelDependencyInjector(ParamsAndPsfModelDependencyInjector):
	"""
	Models the PSF as a 2d gaussian with mean (`x`,`y`), standard deviation `sigma` an offset `const` from zero, and a multiplication factor `factor`
	"""
	
	def __init__(self, psf_data):
		
		super().__init__(psf_data)
		
		self._params = PriorParamSet(
			PriorParam(
				'x',
				(0, psf_data.shape[0]),
				True,
				psf_data.shape[0]//2,
				"Position of gaussian mean on x-axis"
			),
			PriorParam(
				'y',
				(0, psf_data.shape[1]),
				True,
				psf_data.shape[1]//2,
				"Position of gaussian mean on y-axis"
			),
			PriorParam(
				'sigma',
				(0, np.sum([x**2 for x in psf_data.shape])),
				False,
				5,
				"Standard deviation of gaussian in x and y axis"
			),
			PriorParam(
				'const',
				(0, 1),
				False,
				0,
				"Constant added to gaussian"
			),
			PriorParam(
				'factor',
				(0, 2),
				False,
				1,
				"Scaling factor"
			)
		)
		
		self._psf_model = GaussianPSFModel(psf_data.shape, float)
	
	
	
	def get_parameters(self):
		return self._params
	
	def get_psf_model_flattened_callable(self): 
		def psf_model_flattened_callable(x, y, sigma, const, factor):
			return self._psf_model(np.array([x,y]), np.array([sigma,sigma]), const)*factor
		return psf_model_flattened_callable
	
	def get_psf_result_postprocessor(self): 
		return None





"""
Script that deconvolves the first argument with the second argument

Example invocation: 
	`python -m aopp_deconv_tool.deconvolve './example_data/test_rebin.fits{DATA}[10:12]{CELESTIAL:(1,2)}' './example_data/fit_example_psf_000.fits[10:12]{CELESTIAL:(1,2)}'`
"""

import sys, os
from pathlib import Path
from typing import Literal

import numpy as np
from astropy.io import fits

import aopp_deconv_tool.astropy_helper as aph
import aopp_deconv_tool.astropy_helper.fits.specifier
import aopp_deconv_tool.astropy_helper.fits.header
import aopp_deconv_tool.numpy_helper as nph
import aopp_deconv_tool.numpy_helper.axes
import aopp_deconv_tool.numpy_helper.slice
import aopp_deconv_tool.psf_data_ops as psf_data_ops
from aopp_deconv_tool.fpath import FPath
import aopp_deconv_tool.arguments as arguments


from aopp_deconv_tool.algorithm.deconv.clean_modified import CleanModified
from aopp_deconv_tool.algorithm.deconv.lucy_richardson import LucyRichardson

import matplotlib as mpl
import matplotlib.pyplot as plt
import copy
import aopp_deconv_tool.plot_helper as plot_helper
from aopp_deconv_tool.plot_helper.base import AxisDataMapping
from aopp_deconv_tool.plot_helper.plotters import PlotSet, Histogram, VerticalLine, Image, IterativeLineGraph, HorizontalLine

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')


deconv_methods = {
	'clean_modified' : CleanModified,
	'lucy_richardson' : LucyRichardson
}

def create_plot_set(deconvolver, cadence = 1):
	fig, axes = plot_helper.figure_n_subplots(8)
	axes_iter = iter(axes)
	a7_2 = axes[7].twinx()
	
	try:
		cmap = mpl.colormaps['bwr_oob']
	except KeyError:
		cmap = copy.copy(mpl.colormaps['bwr'])
		cmap.set_over('magenta')
		cmap.set_under('green')
		cmap.set_bad('black')
		mpl.cm.register_cmap(name='bwr_oob', cmap=cmap)
	
	try:
		viridis_oob = mpl.colormaps['viridis_oob']
	except KeyError:
		viridis_oob = mpl.colormaps['viridis'].copy()
		viridis_oob.set_bad(color='magenta', alpha=1)
		#viridis_oob.set_under(color='black', alpha=1)
		#viridis_oob.set_over(color='black', alpha=1)
		mpl.cm.register_cmap(name='viridis_oob', cmap=viridis_oob)
	
	
	
	def selected_pixels_non_selected_are_nan(x):
		r = np.array(x._selected_px)
		r[r==0] = np.nan
		return r
	
	plot_set = PlotSet(
		fig,
		'clean modified step={self.n_frames}',
		cadence=cadence,
		plots = [	
			Histogram(
				'residual', 
				static_frame=False,
				axis_data_mappings = (AxisDataMapping('value','bins',limit_getter=plot_helper.lim), AxisDataMapping('count','_hist',limit_getter=plot_helper.LimRememberExtremes()))
			).attach(next(axes_iter), deconvolver, lambda x: x._residual),
		 	
			VerticalLine(
				None, 
				static_frame=False, 
				plt_kwargs={'color':'red'}
			).attach(axes[0], deconvolver, lambda x: x._pixel_threshold),
			
			Image(
		 		'residual'
		 	).attach(next(axes_iter), deconvolver, lambda x: x._residual),
			
			Image(
		 		'current cleaned'
			).attach(next(axes_iter), deconvolver, lambda x: x._current_cleaned),
			
			Image(
		 		'components'
			).attach(next(axes_iter), deconvolver, lambda x: x._components),
			
			Image(
		 		'selected pixels',
				plt_kwargs = {'cmap': viridis_oob},
			).attach(next(axes_iter), deconvolver, lambda x: selected_pixels_non_selected_are_nan(x)),
			
			Image(
		 		'pixel choice metric',
		 		axis_data_mappings = (AxisDataMapping('x',None), AxisDataMapping('y',None), AxisDataMapping('brightness', '_z_data', plot_helper.LimSymAroundValue(0))),
		 		plt_kwargs={'cmap':'bwr_oob'}
			).attach(next(axes_iter), deconvolver, lambda x: x._px_choice_img_ptr.val),
			
			Histogram(
				'pixel choice metric', 
				static_frame=False,
			).attach(next(axes_iter), deconvolver, lambda x: x._px_choice_img_ptr.val),
			
			IterativeLineGraph(
				'metrics',
				datasource_name='fabs',
				axis_labels = (None, 'fabs value (blue)'),
				static_frame=False,
				plt_kwargs = {},
				ax_funcs=[lambda ax: ax.set_yscale('log')]
			).attach(next(axes_iter), deconvolver, lambda x: np.fabs(np.nanmax(x._residual))),
			
			HorizontalLine(
				None, 
				static_frame=False, 
				plt_kwargs={'linestyle':'--'}
			).attach(axes[7], deconvolver, lambda x: x._fabs_threshold),
			
			IterativeLineGraph(
				'metrics',
				datasource_name='rms',
				axis_labels = (None,'rms value (red)'),
				static_frame=False,
				plt_kwargs={'color':'red'},
				ax_funcs=[lambda ax: ax.set_yscale('log')]
			).attach(a7_2, deconvolver, lambda x: np.sqrt(np.nansum(x._residual**2)/x._residual.size)),
			
			HorizontalLine(
				None, 
				static_frame=False, 
				plt_kwargs={'color':'red', 'linestyle':'--'}
			).attach(a7_2, deconvolver, lambda x: x._rms_threshold),
		]
	)
	return plot_set

def run(
		obs_fits_spec : aph.fits.specifier.FitsSpecifier,
		psf_fits_spec : aph.fits.specifier.FitsSpecifier,
		deconvolver : Literal[CleanModified] | Literal[LucyRichardson],
		output_path : str | Path = './deconv.fits',
		plot : bool = True,
	):
	"""
	Given a FitsSpecifier for an observation and a PSF, an output path, and a class that performs deconvolution,
	deconvolves the observation using the PSF.
	
	# ARGUMENTS #
		obs_fits_spec : aph.fits.specifier.FitsSpecifier
			FITS file specifier for observation data, format is PATH{EXT}[SLICE](AXES).
			Where:
				PATH : str
					The path to the FITS file
				EXT : str | int
					The name or number of the FITS extension (defaults to PRIMARY)
				SLICE : "python slice format" (i.e., [1:5, 5:10:2])
					Slice of the FITS extension data to use (defaults to all data)
				AXES : tuple[int,...]
					Axes of the FITS extension that are "spatial" or "celestial" (i.e., RA, DEC),
					by default will try to infer them from the FITS extension header.
		psf_fits_spec : aph.fits.specifier.FitsSpecifier
			FITS file specifier for PSF data, format is same as above
		output_path : str = './deconv.fits'
			Path to output deconvolution to.
		deconvolver : ClassInstance
			Instance of Class to use for deconvolving, defaults to an instance of CleanModified
		plot : bool = True
			If `True` will plot the deconvolution progress
	"""
	
	original_data_type=None
	# Open the fits files
	with fits.open(Path(obs_fits_spec.path)) as obs_hdul, fits.open(Path(psf_fits_spec.path)) as psf_hdul:
		
		# pull out the data we want
		obs_data = obs_hdul[obs_fits_spec.ext].data
		psf_data = psf_hdul[psf_fits_spec.ext].data
		original_data_type=obs_data.dtype
		
		# Create holders for deconvolution products
		deconv_components = np.full_like(obs_data, np.nan)
		deconv_residual = np.full_like(obs_data, np.nan)
		
		# Loop over the index range specified by `obs_fits_spec` and `psf_fits_spec`
		for i, (obs_idx, psf_idx) in enumerate(zip(
				nph.slice.iter_indices(obs_data, obs_fits_spec.slices, obs_fits_spec.axes['CELESTIAL']),
				nph.slice.iter_indices(psf_data, psf_fits_spec.slices, psf_fits_spec.axes['CELESTIAL'])
			)):
			_lgr.debug(f'Operating on slice {i}')
			
			"""
			# Playing with wiener filter deconvolution
			import scipy as sp
			from skimage import restoration
			import scipy.signal
			mask = np.zeros_like(psf_data[psf_idx], dtype=bool)
			mask[tuple(slice(s//4,3*s//4) for s in mask.shape)] = 1
			noise_var = 1E6
			_lgr.debug(f'{noise_var=}')
			wiener_filtered_data = restoration.wiener(obs_data[obs_idx], psf_data[psf_idx], 1E-2, clip=False)
			reconvolve =  sp.signal.fftconvolve(wiener_filtered_data, psf_data[psf_idx], mode='same')
			plt.clf()
			f = plt.gcf()
			ax = f.subplots(2,2).flatten()
			ax[0].imshow(obs_data[obs_idx])
			ax[1].imshow(wiener_filtered_data)
			ax[2].imshow(reconvolve)
			ax[3].imshow(obs_data[obs_idx] -reconvolve)
			plt.show()
			sys.exit()
			"""
		
			# Set up plotting if we want it
			if plot:
				#plt.figure()
				#plt.imshow(obs_data[obs_idx])
				#plt.figure()
				#plt.imshow(psf_data[psf_idx])
				#plt.show()
				plt.close('all')
				plot_set = create_plot_set(deconvolver)
				deconvolver.post_iter_hooks = []
				deconvolver.post_iter_hooks.append(lambda *a, **k: plot_set.update())
				plot_set.show()
			
			# Ensure that we actually have data in this part of the cube
			if np.all(np.isnan(obs_data[obs_idx])) or np.all(np.isnan(psf_data[psf_idx])):
				_lgr.warn('All NAN obs or psf layer detected. Skipping...')
			
			# perform any normalisation and processing
			normed_psf = psf_data_ops.normalise(np.nan_to_num(psf_data[psf_idx]))
			processed_obs = np.nan_to_num(obs_data[obs_idx])
			
			# Store the deconvolution products in the arrays we created earlier
			deconv_components[obs_idx], deconv_residual[obs_idx], deconv_iters = deconvolver(processed_obs, normed_psf)
			
		
		# Save the parameters we used. NOTE: we are only saving the LAST set of parameters as they are all the same
		# in this case. However, if they vary with index they should be recorded with index as well.
		deconv_params = deconvolver.get_parameters()
		
		# Make sure we get all the observaiton header data as well as the deconvolution parameters
		hdr = obs_hdul[obs_fits_spec.ext].header
		hdr.update(aph.fits.header.DictReader(
			{
				'obs_file' : obs_fits_spec.path,
				'psf_file' : psf_fits_spec.path, # record the PSF file we used
				**deconv_params # record the deconvolution parameters we used
			},
			prefix='deconv',
			pkey_count_start=aph.fits.header.DictReader.find_max_pkey_n(hdr)
		))
	
	# Save the deconvolution products to a FITS file
	hdu_components = fits.PrimaryHDU(
		header = hdr,
		data = deconv_components.astype(original_data_type)
	)
	hdu_residual = fits.ImageHDU(
		header = hdr,
		data = deconv_residual.astype(original_data_type),
		name = 'RESIDUAL'
	)
	hdul_output = fits.HDUList([
		hdu_components,
		hdu_residual
	])
	hdul_output.writeto(output_path, overwrite=True)
	
	_lgr.info(f'Deconvolution completed, output written to "{output_path}"')
	

def parse_args(argv):
	import os
	import aopp_deconv_tool.text
	import argparse
	
	DEFAULT_OUTPUT_TAG = '_deconv'
	DESIRED_FITS_AXES = ['CELESTIAL']
	OUTPUT_COLUMNS=80
	try:
		OUTPUT_COLUMNS = os.get_terminal_size().columns - 30
	except Exception:
		pass
	
	FITS_SPECIFIER_HELP = aopp_deconv_tool.text.wrap(
		aph.fits.specifier.get_help(DESIRED_FITS_AXES).replace('\t', '    '),
		OUTPUT_COLUMNS
	)
	DECONV_METHOD_DEFAULT='clean_modified'
	
	class ArgFormatter (argparse.RawTextHelpFormatter, argparse.ArgumentDefaultsHelpFormatter, argparse.MetavarTypeHelpFormatter):
		def __init__(self, *args, **kwargs):
			super().__init__(*args, **kwargs)
	
	
	
	
	
	parser = argparse.ArgumentParser(
		description=__doc__, 
		formatter_class=ArgFormatter,
		epilog=FITS_SPECIFIER_HELP,
		exit_on_error=False
	)
	
	parser.add_argument(
		'obs_fits_spec',
		help = '\n'.join((
			f'The observation\'s (i.e., science target) FITS Specifier. See the end of the help message for more information',
			f'required axes: {", ".join(DESIRED_FITS_AXES)}',
		)),
		type=str,
		metavar='FITS Specifier',
	)
	
	parser.add_argument(
		'psf_fits_spec',
		help = '\n'.join((
			f'The psf\'s (i.e., calibration target) FITS Specifier. See the end of the help message for more information',
			f'required axes: {", ".join(DESIRED_FITS_AXES)}',
		)),
		type=str,
		metavar='FITS Specifier',
	)
	
	parser.add_argument(
		'-o', 
		'--output_path', 
		type=FPath,
		metavar='str',
		default='{parent}/{stem}{tag}{suffix}',
		help = '\n    '.join((
			f'Output fits file path, supports keyword substitution using parts of `obs_fits_spec` path where:',
			'{parent}: containing folder',
			'{stem}  : filename (not including final extension)',
			f'{{tag}}   : script specific tag, "{DEFAULT_OUTPUT_TAG}" in this case',
			'{suffix}: final extension (everything after the last ".")',
			'\b'
		))
	)
	parser.add_argument(
		'--plot', 
		action='store_true', 
		default=False, 
		help='If present will show progress plots of the deconvolution'
	)
	parser.add_argument(
		'--deconv_method', 
		type=str, 
		choices=deconv_methods.keys(), 
		default=DECONV_METHOD_DEFAULT, 
		help='Which method to use for deconvolution. For more information, pass the deconvolution method and the "--info" argument.'
	) 
	parser.add_argument(
		'--deconv_method_help', 
		action='store_true', 
		default=False, 
		help='Show help for the selected deconvolution method'
	)
	
	parser.successful = True
	parser.error_message = None
	
	def on_error(err_str):
		parser.successful = False
		parser.error_message = err_str
		
	parser.error = on_error
	
	args, deconv_args = parser.parse_known_args(argv)
	
	if not parser.successful:
		print(vars(args))
		if args.deconv_method_help:
			return args, deconv_args
		else:
			parser.print_usage()
			print(parser_error_message)
			sys.exit()
			
	
	
	args.obs_fits_spec = aph.fits.specifier.parse(args.obs_fits_spec, DESIRED_FITS_AXES)
	args.psf_fits_spec = aph.fits.specifier.parse(args.psf_fits_spec, DESIRED_FITS_AXES)
	
	other_file_path = Path(args.obs_fits_spec.path)
	args.output_path = args.output_path.with_fields(
		tag=DEFAULT_OUTPUT_TAG, 
		parent=other_file_path.parent, 
		stem=other_file_path.stem, 
		suffix=other_file_path.suffix
	)
	
	return args, deconv_args


def go(
		obs_fits_spec,
		psf_fits_spec,
		output_path=None, 
		plot=None, 
		deconv_method=None, 
		deconv_method_help_FLAG=None, 
		**kwargs
	):
	"""
	Thin wrapper around `run()` to accept string inputs.
	As long as the names of the arguments to this function 
	are the same as the names expected from the command line
	we can do this programatically
	"""
	# This must be first so we only grab the arguments to the function
	fargs = dict(locals().items())
	arglist = aopp_deconv_tool.arguments.construct_arglist_from_locals(fargs, n_positional_args=2)
	
	exec_with_args(arglist)
	return

def exec_with_args(argv):
	args, deconv_args = parse_args(argv)
	_lgr.debug('#### ARGUMENTS ####')
	for k,v in vars(args).items():
		_lgr.debug(f'\t{k} : {v}')
	_lgr.debug('#### END ARGUMENTS ####')
	
		
	deconv_class = deconv_methods[args.deconv_method]
	deconv_params = arguments.parse_args_of_dataclass(
		deconv_class, 
		deconv_args, 
		prog=f'deconvolve.py --deconv_method {args.deconv_method}',
		show_help=args.deconv_method_help
	)
	
	_lgr.debug(f'{deconv_params=}')
	deconvolver = deconv_class(**deconv_params)
	
	run(
		args.obs_fits_spec, 
		args.psf_fits_spec, 
		deconvolver = deconvolver,
		output_path = args.output_path, 
		plot = args.plot,
	)
	
	return
	
if __name__ == '__main__':
	argv = sys.argv[1:]
	exec_with_args(argv)
	





from typing import Callable, TypeVar

import numpy as np
import scipy as sp

import matplotlib.pyplot as plt

type N = TypeVar('N',bound=int)
type N_plus_one = TypeVar('N',bound=int)


def difference_of_scale_filters(
		data : np.ndarray, 
		lo_scale : int = 0, 
		hi_scale : int = 1, 
		filter : Callable[[np.ndarray, int], np.ndarray] = sp.ndimage.uniform_filter,
		**kwargs
	) -> np.ndarray:
	print(f'{lo_scale=} {hi_scale=}')
	if lo_scale == 0:
		lo_data = data
	else:
		lo_data = filter(data, lo_scale, **kwargs)
	return lo_data - filter(data, hi_scale, **kwargs)

def wavelet_decomposition(
		data : np.ndarray['N'],
		lo_scale : int = 1,
		hi_scale : int | None = None,
		filter : Callable[[np.ndarray, float|int], np.ndarray] = sp.ndimage.uniform_filter,
		delta_function_scale = 1,
		ensure_sizes_are_odd=True,
		**kwargs
	) -> np.ndarray['N_plus_one']:
	
	hi_scale = min(int(np.floor(np.log2(s))) for s in data.shape) if hi_scale is None else hi_scale
	wavelet_planes = np.zeros((hi_scale - lo_scale+2, *data.shape), dtype=data.dtype)
	data_decumulator = np.array(data)
	
	for i, scale in enumerate(range(lo_scale, hi_scale)):
		lo_filter_is_delta_function = scale == delta_function_scale
		
		lo_size = 2**(scale-1) + (int(ensure_sizes_are_odd) if ((scale-1)>0) else 0)
		hi_size = 2**(scale) + (int(ensure_sizes_are_odd) if ((scale)>0) else 0)
		print(f'{i=} {lo_size=} {hi_size=} {lo_filter_is_delta_function=}')
		
		norm = np.nansum(data_decumulator)
		lo_filtered = data if lo_filter_is_delta_function else filter(data, lo_size)
		lo_filtered *= (np.nansum(lo_filtered)/norm)
		hi_filtered = filter(data, hi_size)
		hi_filtered *= (np.nansum(hi_filtered)/norm)
		
		wavelet_planes[i] = lo_filtered - hi_filtered
		
		data_decumulator = data - np.nansum(wavelet_planes[:i+1], axis=0)
	
	print(f'last {i=}')
	wavelet_planes[i+1] = filter(data, 2**hi_scale+int(ensure_sizes_are_odd))
	data_decumulator = data - np.nansum(wavelet_planes[:i+2], axis=0)
	wavelet_planes[i+2] = data_decumulator
	
	return wavelet_planes


"""
Point spread function (PSF) model that is just a gaussian
"""

from typing import Any, TypeVar, TypeVarTuple, Generic, Annotated, NewType

import numpy as np

IntVar = TypeVar('IntVar', bound=int)
T = TypeVar('T')
U = TypeVar('T')
Ts = TypeVarTuple('Ts')
class S(Generic[IntVar]): pass
class S1(Generic[IntVar]): pass
N = TypeVar('N',bound=int)

import matplotlib.pyplot as plt

class GaussianPSFModel:
	"""
	Class defines a gaussian where the shape and type are fixed at instantiaion, but the mean, standard deviation, and offset are calculated
	by the instance. This is useful for putting into an optimising function as the underlying array storage is only created once.
	"""
	def __init__(self,
			shape : S[N],
			dtype : T
		):
		self.result = np.zeros(shape, dtype=dtype)
		self.indices = np.indices(self.result.shape)
		self.working_array = np.zeros(self.indices.shape, dtype=dtype)
	
	def __call__(self,
			mean : np.ndarray[S1[N],U],
			std : np.ndarray[S1[N],U],
			const : U
		) -> np.ndarray[S[N],T]:
		
		# const from -1 -> 1, where 1 is the amplitude of the gaussian, but still want everything to sum to 1
		inverse_a = 1 - const
		
		
		self.working_array = self.indices.T - mean
		
		norm_factor = np.sqrt(((2*np.pi)**std.size)) * np.prod(std)
		self.working_array = (self.working_array**2)/(2*(std**2))
		
		self.result[...] =  inverse_a * 1/norm_factor * np.exp(-np.sum(self.working_array, axis=-1)).T + const / self.result.size
		
		return self.result


"""
Contains routines for casting to and from types. Generally useful for simple data marshalling to and from text
"""

from typing import Iterable, Any

class CastException(Exception):
	"""
	If casts fail, throw this exception
	"""
	pass


def to(x : Any, atype : type) -> type:
	"""
	Casts an object to a type
	
	Arguments:
		x : any
			Object to cast
		atype : type
			type to cast to
	
	Returns:
		x casted to `atype`, throws CastException on failure.
	"""
	try:
		return atype(x)
	except:
		raise CastException(f"Cannot cast type {type(x)} = '{x}' to type {atype}")

def to_any(x : Any, types : Iterable[type]) -> Any:
	"""
	Casts and object to any of the given types.
	
	Arguments:
		x : Any
			Object to cast
		types : Iterable[type]
			Iterable (e.g., list) of the types to attempt to cast to. Once one
			cast succeeds, the others are not attempted.
	
	Returns:
		`x` casted to the first type that succeeds. Throws CastException on failure.
	"""
	for atype in types: 
		try:
			return(to(x, atype))
		except CastException:
			continue
	raise CastException(f"Cannot cast type {type(x)} = '{x}' to any of the types {tuple(t for t in types)}")




from aopp_deconv_tool.psf_model.dependency_injector.params_and_psf_model_di import *
from aopp_deconv_tool.psf_model.turbulence_psf_model import TurbulencePSFModel


class TurbulencePSFModelDependencyInjector(ParamsAndPsfModelDependencyInjector):
	"""
	Models the PSF as the result of von-karman turbulence. Assumes the PSF is at the centre of the model.
	"""
	
	def __init__(self, psf_data):
		super().__init__(psf_data)
		
		self._params = PriorParamSet(
			PriorParam(
				'wavelength',
				(0, np.inf),
				True,
				750E-9,
				"Wavelength (meters) that properties are calculated at, will be automatically set if spectral information is present in the FITS cube"
			),
			PriorParam(
				'r0',
				(0, 1),
				True,
				0.1,
				"Fried Parameter"
			),
			PriorParam(
				'turbulence_ndim',
				(0, 3),
				False,
				1.5,
				"Number of dimensions the turbulence has"
			),
			PriorParam(
				'L0',
				(0, 50),
				False,
				8,
				"Von-Karman turbulence parameter"
			)
		)
		
		self._psf_model = TurbulencePSFModel(
			SimpleTelescope(
				8, 
				200, 
				CCDSensor.from_shape_and_pixel_size(psf_data.shape, 2.5E-6)
			),
			turbulence_model
		)
		
	def get_parameters(self):
		return self._params
	
	def get_psf_model_flattened_callable(self): 
		return self._psf_model
	
	def get_psf_result_postprocessor(self): 
		return None





from aopp_deconv_tool.optics.geometric.optical_component import OpticalComponentSet
from aopp_deconv_tool.optics.telescope_model import optical_transfer_function_of_optical_component_set


class InstrumentBase:
	"""
	Interface for instrument descriptions to be used with the PSFModel class.
	"""
	f_ao : float | None = None # adaptive optics spatial freq (if required)
	ocs : OpticalComponentSet | None = None # optical component set that describes the instrument
	
	def __init__(self,
			obs_shape,
			obs_scale,
			obs_pixel_size,
			ref_wavelength,
			expansion_factor : float, 
			supersample_factor : float
		):
		self.obs_shape = obs_shape
		self.obs_scale = obs_scale
		self.obs_pixel_size = obs_pixel_size
		self.ref_wavelength = ref_wavelength
		self.expansion_factor = expansion_factor
		self.supersample_factor= supersample_factor
	
	def optical_transfer_function(self):
		
		return optical_transfer_function_of_optical_component_set(
			self.obs_shape, 
			self.expansion_factor, 
			self.supersample_factor, 
			self.ocs, 
			self.obs_scale
		)




class BiDirectionalMap:
	"""
	A map that can efficiently go from A->B and B->A.
	"""
	def __init__(self, seed_dict={}):
		self.is_forward = True
		self.forward_dict = dict((k,v) for k,v in seed_dict.items())
		self.backward_dict = dict((v,k) for k,v in seed_dict.items())
	
	def __getattr__(self, name):
		proxy = self.forward_dict if self.is_forward else self.backward_dict
		if hasattr(proxy, name):
			return getattr(proxy, name)
		else:
			raise AttributeError(f'No attribute {name} on {type(self).__name__}')
	
	@property
	def backward(self):
		self.is_forward = False
		return self
	
	def __getitem__(self, key):
		if self.is_forward:
			return self.forward_dict[key]
		else:
			self.is_forward = True
			return self.backward_dict[key]
	
	def __contains__(self, key):
		if self.is_forward:
			return key in self.forward_dict
		else:
			self.is_forward = True
			return key in self.backward_dict
	
	def __setitem__(self, key, value):
		if self.is_forward:
			self.forward_dict.__setitem__(key,value)
			self.backward_dict.__setitem__(value,key)
		else:
			self.is_forward = True
			self.forward_dict.__setitem__(value,key)
			self.backward_dict.__setitem__(key,value)
		return
	
	def update(self, *args, **kwargs):
		is_forward_flag = self.is_forward
		for k,v in dict(*args, **kwargs).items():
			self.is_forward = is_forward_flag
			self.__setitem__(k,v)
	
	def __repr__(self):
		return f'{type(self).__name__}({str(self.forward_dict)})'



"""
Contains type definitions of optical quantities. Useful for static type checking
and making sure that we are sending the correct data somewhere.
"""

import numpy as np

from aopp_deconv_tool.optics.geometric.optical_component import OpticalComponentSet
from aopp_deconv_tool.geo_array import GeoArray

import aopp_deconv_tool.cfg.logs

_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')


class PupilFunction(GeoArray):
	"""
	Represents the pupil function of an optical system
	"""

	

class PointSpreadFunction(GeoArray):
	"""
	Represents the point spread function of an optical system
	"""
	@classmethod
	def from_pupil_function(cls,
			  pupil_function : PupilFunction
		):
		"""
		Calculate the PSF from a pupil function
		"""
		pf_fft = pupil_function.fft()
		return cls(
			np.conj(pf_fft.data)*pf_fft.data,
			pf_fft.axes,
		)

	@classmethod
	def from_optical_transfer_function(cls,
			optical_transfer_function,
		):
		"""
		Calculate the PSF from the optical transfer function of an optical system
		"""
		otf_ifft = optical_transfer_function.ifft()
		
		return cls(
			otf_ifft.data,
			otf_ifft.axes,
		)
		

class OpticalTransferFunction(GeoArray):
	@classmethod
	def from_psf(cls, point_spread_function : PointSpreadFunction):
		"""
		Get the optical transfer function from a point_spread_function of an optical system
		"""
		psf_fft = point_spread_function.fft()
		return cls(
			psf_fft.data,
			psf_fft.axes,#/wavelength
		)


class PhasePowerSpectralDensity(GeoArray):
	"""
	Represents the phase power spectral density of an optical system
	"""
	





"""
Contains routines that are mostly wrappers around scipy interpolation functions
"""
from typing import Literal, Any
import numpy as np
import scipy as sp
import scipy.interpolate

def interpolate_at_mask(data : np.ndarray[[...], Any], mask : np.ndarray[[...],bool], edges : None | Literal['convolution'] = None, **kwargs) -> np.ndarray[[...],Any]:
	"""
	Interpolates an array 'data' at the True points given in 'mask'
	**kwargs is passed through to "sp.interpolate.griddata()"
	"""
	if edges is None:
		interp_data = np.array(data)
	elif edges == 'convolution':
		# build convolution kernel
		kernel_shape = tuple(3 for s in data.shape)
		kernel_edge_shape = tuple(s//2 for s in kernel_shape)
		kernel = np.ones(kernel_shape)
		kernel /= np.sum(kernel)
		embed_shape = tuple(s+2*kes for kes, s in zip(kernel_edge_shape, data.shape))
		embed_slice = tuple(slice(kes,-kes) for kes in kernel_edge_shape)
		
		# build data and mask with a frame of zeros
		interp_data = np.zeros(embed_shape)
		interp_data[embed_slice] = data
		interp_mask = np.zeros(embed_shape, dtype=mask.dtype)
		interp_mask[embed_slice] = mask
		mask = interp_mask
		
		# interpolate the framed data, this should remove the effects of any masked out points
		interp_data = interpolate_at_mask(interp_data, interp_mask, **kwargs)
		# convolve the interpolated framed data, the frame should now be non-zero but tending towards zero
		interp_data = sp.signal.fftconvolve(interp_data, kernel, mode='same')
		# put the original data back into the frame, we are just using the frame to anchor the interpolation.
		interp_data[embed_slice] = data
	else:
		raise ValueError(f'Unknown edge interpolation strategy "{edges=}"')

	#points = np.array(mgrid_from_array(interp_data, gridder=np.mgrid))
	points = np.indices(interp_data.shape)

	p_known = points[:,~mask].T
	p_unknown = points[:,mask].T
	known_values = interp_data[~mask]

	interp_values = sp.interpolate.griddata(p_known, known_values, p_unknown, **kwargs)

	interp_data[mask] = interp_values
	
	if edges is None:
		return(interp_data)
	elif edges == 'convolution':
		return(interp_data[embed_slice]) # only want the embedded interpolated data, not the frame
	raise ValueError(f'Unknown edge interpolation strategy "{edges=}", this should never happen.')


"""
Routines for identifying and fixing bad pixels in observation data.
"""
import aopp_deconv_tool.numpy_helper as nph
import aopp_deconv_tool.numpy_helper.array.mask
import aopp_deconv_tool.numpy_helper.array.mask.interp

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'INFO')

def get_map(a):
	return nph.array.mask.from_nan_and_inf(a)

def fix(a, bp_map, method, window=1,boundary='reflect',const=0):
	match method:
		case 'simple':
			return nph.array.mask.interp.constant(a,bp_map,const)
		case 'mean':
			return nph.array.mask.interp.mean(a,bp_map,window,boundary,const)
		case 'interp':
			return nph.array.mask.interp.interp(a,bp_map)
		case _:
			raise RuntimeError(f'Unknown case {method}')
			
	
